#include "alignment.h"
#include "run.h"

#include <math.h>
/* Symbols used for displaying the alignment */
#define DISPLAYED_MATCH_SYMBOL '|'
#define DISPLAYED_MISMATCH_SYMBOL '.'
#define DISPLAYED_SEQ_GAP_SYMBOL '-'
#define DISPLAYED_MATCH_GAP_SYMBOL ' '


/* Encoding/Decoding of the 'from' field */
#define I_MASK 0x2
#define J_MASK 0x1
#define FROM_LEFT(direction)   (!((direction) & I_MASK) &&  ((direction) & J_MASK))
#define FROM_TOP(direction)    ( ((direction) & I_MASK) && !((direction) & J_MASK))
#define FROM_CORNER(direction) ( ((direction) & I_MASK) &&  ((direction) & J_MASK))
#define D_I(direction) (-(((direction) & I_MASK) >> 1))
#define D_J(direction) (-( (direction) & J_MASK))

#define COLOR_MISMATCH_PERSISTENCE(q_level) ((q_level)) /*(((q_level) + 1) >> 1)*/


/* 'Shortcuts' for some frequently used code blocks */
#define CHECK_RELATIVE_COORDS(i, j) ((i) >= 0 && (i) < alignment->read_len && (j) >= alignment->range[i][ALIGNMENT_REFERENCE_START] && (j) - alignment->range[i][ALIGNMENT_REFERENCE_START] < alignment->range[i][ALIGNMENT_LINE_LEN])

#define CHECK_ABSOLUTE_COORDS(i, j) ((i) >= 0 && (i) < alignment->read_len && (j) >= 0 && (j) < alignment->range[i][ALIGNMENT_LINE_LEN])

#define REF_START(i) (alignment->range[i][ALIGNMENT_REFERENCE_START])
#define REF_END(i) (alignment->range[i][ALIGNMENT_REFERENCE_START] + alignment->range[i][ALIGNMENT_LINE_LEN])
#define REF_LINE_LEN(i) (alignment->range[i][ALIGNMENT_LINE_LEN])

#define TO_MATRX_ABS_COORD(j, i) ((j) + 1 - alignment->range[i][ALIGNMENT_REFERENCE_START])
#define TO_REFERENCE_COORD(j, i) ((j) - 1 + alignment->range[i][ALIGNMENT_REFERENCE_START])



ScoreType min_accepted_score = MIN_ACCEPTED_SCORE;
ScoreType min_accepted_score_simd_filter = MIN_ACCEPTED_SCORE_SIMD_FILTER;
ScoreType min_trusted_score = MIN_TRUSTED_SCORE;

inline void alignment__create_martix(AlignmentType * alignment, int allowed_indels);

inline void alignment__init_read_quality(AlignmentType * alignment, short * read_quality);


inline ScoreType alignment__score_for_match(AlignmentType * alignment, CODE_TYPE color_match, short quality_level);

inline ScoreType alignment__score_for_pair(AlignmentType * alignment, CODE_TYPE color_1, CODE_TYPE color_2, short quality_level);


inline AlignmentCellType* alignment__matrix_access(AlignmentType * alignment, int i, int j);

inline AlignmentCellType* alignment__matrix_access__no_check(AlignmentType * alignment, int i, int j);

inline ScoreType alignment__get_score__no_check(const AlignmentType * alignment, int i, int j);


/**
 * Alignment structures creation and initialization, for just one use (read and reference are already provided)
 * @param read Read sequence
 * @param read_quality Read sequence quality
 * @param read_len Read length
 * @param reference Reference sequence, to which the read is aligned
 * @param read_len Reference sequence length
 * @param allowed_indels Max number of allowed indels in the alignment; usually small, it prevents from computing the entire DP matrix, thus the reducing computation time.
 * @return The address of the alignment data structure
 */
AlignmentType* alignment__create1(CODE_TYPE* read, short* read_quality, short read_len, CODE_TYPE* reference, short ref_len, short allowed_indels) {
  AlignmentType* alignment = alignment__create(read_len, ref_len, allowed_indels);
  alignment__reset(alignment, read, read_quality, reference);
  return alignment;
}

/**
 * Alignment structures creation and initialization
 * @param read_len Read length
 * @param read_len Reference sequence length
 * @param allowed_indels Max number of allowed indels in the alignment; usually small, it prevents from computing the entire DP matrix, thus the reducing computation time.
 * @return The address of the alignment data structure
 */
AlignmentType* alignment__create(short read_len, short ref_len, short allowed_indels) {
  /* what is constructed and "returned" */
  AlignmentType * alignment;
  SAFE_FAILURE__ALLOC(alignment, 1, AlignmentType);
  alignment__init(alignment, read_len, ref_len, allowed_indels);
  return alignment;
}

/**
 * Alignment structures creation and initialization
 * @param alignment
 * @param read_len Read length
 * @param read_len Reference sequence length
 * @param allowed_indels Max number of allowed indels in the alignment; usually small, it prevents from computing the entire DP matrix, thus the reducing computation time.
 */
inline void alignment__init(AlignmentType* alignment, short read_len, short ref_len, short allowed_indels) {
  /* what is constructed and "returned" */
  alignment->params.allowed_indels = allowed_indels;

  alignment__set_params(alignment, MATCH, MISMATCH, GAP_OPEN, GAP_EXTEND);

  SAFE_FAILURE__ALLOC(alignment->read, (size_t)read_len, CODE_TYPE);
  alignment->read_len = read_len;

  SAFE_FAILURE__ALLOC(alignment->reference, (size_t)ref_len, CODE_TYPE);
  alignment->ref_len = ref_len;

  SAFE_FAILURE__ALLOC(alignment->read_quality, (size_t)read_len, ScoreType);

  /* The size of the reference sequence indicates the number of indels accepted; not sure if this is a good idea; TODO to be reviewed */
  /* Reference is considered larger than the read; */
  alignment__create_martix(alignment, allowed_indels);

  /* Create the string where the alignment is displayed;
     3 times (the worst case alignment length + 1 for the new line) + 1 for the \0 */
  SAFE_FAILURE__ALLOC(alignment->to_display, 3*(read_len + ref_len + 1) + 1, char);
  SAFE_FAILURE__ALLOC(alignment->traceback, read_len + allowed_indels, unsigned char);
  SAFE_FAILURE__ALLOC(alignment->read_indel_positions, alignment->params.allowed_indels, short);
  alignment->status = UNALIGNED;
}

/**
 * Actual construction of the alignment "matrix", more precisely a part of it.
 * The read sequence symbols label the matrix lines, while the reference sequence symbols index the columns. On each line, only some columns are actually allocated: because of the small number of allowed indels, each symbol in the read has a limited number of allowed pairs in the reference; their range is memorized in the 'range' vector.
 * @param alignment The address of the alignment data structures
 * @param allowed_indels The maximum number of indels allowed in the alignment.
 */
inline void alignment__create_martix(AlignmentType * alignment, int allowed_indels) {
  int mat_k;
  int i;
  int mat_size = alignment->read_len;
  /* Few indels allowed: no need to compute the entire DP matrix, just a neighborhood of the diagonal */
  int seq_dif = (alignment->ref_len - alignment->read_len + 1);
  int max_line_len = min(alignment->ref_len, max((2*allowed_indels) + 1, seq_dif));

  /* Create the alignment matrix */
  /* consider the reference on top and the read on the bottom (side))*/
  SAFE_FAILURE__ALLOC(alignment->matrix, alignment->read_len, AlignmentCellType**);
  SAFE_FAILURE__ALLOC(alignment->range, mat_size, short*);
  for (mat_k = 0; mat_k < mat_size; ++mat_k) {
	SAFE_FAILURE__ALLOC(alignment->range[mat_k], ALIGNMENT_RANGE_INFO_SIZE, short);
    /* where the overlap starts in the reference sequence */
    alignment->range[mat_k][ALIGNMENT_REFERENCE_START] = max(0, mat_k - (allowed_indels ? (allowed_indels - 1) : 0));
    alignment->range[mat_k][ALIGNMENT_LINE_LEN] =
      max_line_len -
      max(0, allowed_indels - mat_k - 1) -
      max(0, (max_line_len +
      alignment->range[mat_k][ALIGNMENT_REFERENCE_START]) - alignment->ref_len);
    //alignment->matrix[mat_k] = (AlignmentCellType**)calloc(/*alignment->range[mat_k][ALIGNMENT_LINE_LEN]*/ max_line_len + 2, sizeof(AlignmentCellType*));
	SAFE_FAILURE__ALLOC(alignment->matrix[mat_k], max_line_len + 2, AlignmentCellType*);
    /* each line has 2 extra cells, one  on the left and one on the right, each holding the minimum possible score,
     * with the sole purpose of allowing the same recurrence formula for all cells, including those on the borders */

    for (i = 0; i < alignment->range[mat_k][ALIGNMENT_LINE_LEN]+2; ++i) {
      SAFE_FAILURE__ALLOC(alignment->matrix[mat_k][i], 1, AlignmentCellType);
    }
  // 'fake' left border cell, that never changes
  (alignment->matrix[mat_k][0])->score = MIN_SCORE;
  (alignment->matrix[mat_k][0])->mismatches = MIN_SCORE;
  // 'fake' right border cell, that never changes
  (alignment->matrix[mat_k][alignment->range[mat_k][ALIGNMENT_LINE_LEN] + 1])->score = MIN_SCORE;
  (alignment->matrix[mat_k][alignment->range[mat_k][ALIGNMENT_LINE_LEN] + 1])->mismatches = MIN_SCORE;
  // the color checksum is already 0 (identity)
  // the direction masks and indel count are null
  }
  //alignment__init_martix(alignment);
}

/**
 * Initialize the alignment matrix
 * @param alignment The address of the alignment data structures
 */
inline void alignment__init_martix(AlignmentType * alignment) {
  int mat_k;
  int j;
  int mat_size = alignment->read_len;
  for (mat_k = 0; mat_k < mat_size; ++mat_k) {
    for (j = 1; j <= alignment->range[mat_k][ALIGNMENT_LINE_LEN]; ++j) {
      (alignment->matrix[mat_k][j])->color_checksum =
        COMPOSE(alignment->read[mat_k], alignment->reference[TO_REFERENCE_COORD(j, mat_k)]);
      /*Score initialized with the match / mismatch score adjusted wrt the quality */
      (alignment->matrix[mat_k][j])->score =
        alignment__score_for_match(alignment, alignment->matrix[mat_k][j]->color_checksum, alignment->read_quality[mat_k]);
      (alignment->matrix[mat_k][j])->from = I_MASK | J_MASK;
      (alignment->matrix[mat_k][j])->indels = 0;
      (alignment->matrix[mat_k][j])->color_autocorrect = (alignment->matrix[mat_k][j])->color_checksum;
      (alignment->matrix[mat_k][j])->color_mismatch_freshness = ((alignment->matrix[mat_k][j])->color_autocorrect) ? COLOR_MISMATCH_PERSISTENCE(alignment->read_quality[mat_k]) : COLOR_MISMATCH_NONE;
    }
  }
  alignment->best_i0 = alignment->best_j0 = alignment->best_i = alignment->best_j = -1;
}
/**
 * Initialize the read_quality vector
 * The memory is considered allocated, and all the sizes correct (i.e. read_len)
 * @param alignment
 * @param read_quality The read quality 'classes' (levels), short integers ranging from 0 to READ_QUALITY_LEVELS
 */
inline void alignment__init_read_quality(AlignmentType * alignment, short* read_quality) {
  int i;
  if (read_quality != NULL) {
    for (i = 0; i <  alignment->read_len; ++i) {
      alignment->read_quality[i] = QUALITY_LEVEL(min(MAX_READ_QUALITY, read_quality[i]));
    }
  } else {
    for (i = 0; i <  alignment->read_len; ++i) {
      alignment->read_quality[i] = QUALITY_LEVEL(MAX_READ_QUALITY);
    }
  }
}

/**
 * Initialize the read_quality vector, from a compressed source
 * The memory is considered allocated, and all the sizes correct (i.e. read_len)
 * @param alignment
 * @param read_quality The read quality 'classes' (levels), short integers ranging from 0 to READ_QUALITY_LEVELS
 */
inline void alignment__init_read_quality_compressed(AlignmentType * alignment, QUAL_TYPE* read_quality) {
  int i;
  if (read_quality != NULL) {
    for (i = 0; i <  alignment->read_len; ++i) {
      alignment->read_quality[i] = NTH_QUAL(read_quality, i);
    }
  } else {
    for (i = 0; i <  alignment->read_len; ++i) {
      alignment->read_quality[i] = QUALITY_LEVEL(MAX_READ_QUALITY);
    }
  }
}

/**
 * Reuse the alignment data structure, when the read and reference have the same lengths as the previous ones aligned.
 * @param alignment The address of the data structure to reuse
 * @param read The new read sequence, known to have the length alignment->read_len
 * @param read_quality The new read quality sequence, known to have the length alignment->read_len
 * @param reference The new reference sequence, known to have the length alignment->ref_len
 */
inline void alignment__reset(AlignmentType *alignment, CODE_TYPE* read, short* read_quality, CODE_TYPE* reference){
  memcpy(alignment->read, read, alignment->read_len * sizeof(CODE_TYPE));
  memcpy(alignment->reference, reference, alignment->ref_len * sizeof(CODE_TYPE));
  alignment__init_read_quality(alignment, read_quality);
  memset(alignment->read_indel_positions, 0, alignment->params.allowed_indels * sizeof(short));
   alignment->status = UNALIGNED;
  //alignment__init_martix(alignment);
}
/**
 * Reuse the alignment data structure, when the read and reference have the same lengths as the previous ones aligned, and come from compressed sources.
 * @param alignment The address of the data structure to reuse
 * @param read The new read sequence, known to have the length alignment->read_len
 * @param read_quality The new read quality sequence, known to have the length alignment->read_len
 * @param reference The new reference sequence, known to have the length alignment->ref_len
 * @param offset How many codes to skip in the compressed reference
 */
inline void alignment__reset_with_compressed(AlignmentType *alignment, CODE_TYPE* read, QUAL_TYPE* read_quality, CODE_TYPE* reference, int ref_offset){
  decompress_to(read,      alignment->read,      alignment->read_len, 0);
  decompress_to(reference, alignment->reference, alignment->ref_len,  ref_offset);
  alignment__init_read_quality_compressed(alignment, read_quality);
  memset(alignment->read_indel_positions, 0, alignment->params.allowed_indels * sizeof(short));
   alignment->status = UNALIGNED;
  //alignment__init_martix(alignment);
}

/**
 * Reuse the alignment data structure, for a new read that has the same length as the previous one, and comes from a compressed source.
 * the matrix is not initialized, as the new reference is also ewpected.
 * @param alignment The address of the data structure to reuse
 * @param reference The new reference sequence, known to have the length alignment->ref_len
 * @param offset How many codes to skip in the compressed reference
 */
inline void alignment__reset_with_compressed_read(AlignmentType *alignment,CODE_TYPE* read, QUAL_TYPE* read_quality) {
  decompress_to(read,      alignment->read,      alignment->read_len, 0);
  alignment__init_read_quality_compressed(alignment, read_quality);
  memset(alignment->read_indel_positions, 0, alignment->params.allowed_indels * sizeof(short));
  //alignment__init_martix(alignment);
}

/**
 * Reuse the alignment data structure, when the read stays the same and reference has the same lengths as the previous one, and comes from a compressed source.
 * @param alignment The address of the data structure to reuse
 * @param reference The new reference sequence, known to have the length alignment->ref_len
 * @param offset How many codes to skip in the compressed reference
 */
inline void alignment__reset_with_compressed_ref(AlignmentType *alignment, CODE_TYPE* reference, int ref_offset){
  decompress_to(reference, alignment->reference, alignment->ref_len,  ref_offset);
   alignment->status = UNALIGNED;
  //alignment__init_martix(alignment);
}
/**
 * Frees the memory of an alignment
 * @param alignment The address of the data structure to be disposed of
 */
void alignment__destroy(AlignmentType *alignment) {
  int i, j;
  free(alignment->read);
  free(alignment->reference);
  free(alignment->read_quality);
  free(alignment->read_indel_positions);
  for (i = 0; i < alignment->read_len; ++i) {
    for (j = 0; j < alignment->range[i][ALIGNMENT_LINE_LEN] + 2; ++j) {
      free(alignment->matrix[i][j]);
    }
    free(alignment->matrix[i]);
    free(alignment->range[i]);
  }
  free(alignment->matrix);
  free(alignment->range);
  if (alignment->to_display) {
    free(alignment->to_display);
  }
  if (alignment->traceback) {
    free(alignment->traceback);
  }
}

/**
 * Alignment algorithm parameter settings.
 * @param alignment The address of the alignment data structure
 * @param match The match score
 * @param mismatch The mismatch penalty
 * @param gap_open The penalty for gap opening
 * @param gap_extend The penalty for gap extension
 */
void alignment__set_params(AlignmentType * alignment, short match, short mismatch, short gap_open, short gap_extend) {
  int i;
  alignment->params.gap[GAP_OPEN_IDX]   = gap_open;
  alignment->params.gap[GAP_EXTEND_IDX] = gap_extend;
  /*
  * Pre-compute match/mismatch scores for several quality intervals.
  * Phred quality, log transformed error probability: Q = -10 log10(P), where P is the error probability
  * (http://www.phrap.org/phredphrapconsed.html)
  * => P = 10^-(Q/10)
  *
  * In the pair_scores table, the second index (i) has the following relation to the quality value:
  * (READ_QUALITY_LEVELS - 1) * min (Q, 40) / 40 => index of the score
  * (beyond Q=40, the error probability is less than 0.0001, and is considered insignificant)
  *
  * For match, the lower is the quality, the lower is the reward:
  * i / (float)READ_QUALITY_LEVELS * matchscore
  * Similarly, For mismatch, the lower is the quality, the lower is the mismatch penalty:
  * i / (float)READ_QUALITY_LEVELS * mismatchpenalty
  *
  */
  for (i = 0; i < READ_QUALITY_LEVELS; ++i) {
    alignment->params.pair_scores[SYMBOL_PAIR_MATCH_IDX][i]    = SCALE_SCORE(   match, i);
    alignment->params.pair_scores[SYMBOL_PAIR_MISMATCH_IDX][i] = SCALE_SCORE(mismatch, i);
  }
  /* should be followed by __init_matrix for proper effect */
  //if (alignment->matrix) {
  //  alignment__init_martix(alignment);
  //}
}

/**
 * In an alignment, if the composition of a series of N color mismatches becomes CODE_IDENTITY, then it corresponds to N-1 base mismatches
 * in the DNA space, and should be penalized less. A color mismatch is not penalized (gets score 0) if it corrects such
 * a series of color mismatches.
 * @param prev Previous code composition
 * @param crt Current match code
 * @param crt_score The score that would normally be given to the symbol pair, regardless of color corrections.
 * @return The score that should be given to the symbol pair wrt the eventual color correction:
 * O (no penalty) if there is a correction, or the previously established score otherwise.
 */
inline ScoreType alignment__color_correction_score_adjustment(CODE_TYPE prev, CODE_TYPE crt, CODE_TYPE reference, ScoreType crt_score) {
  //TODO improve
  return (prev && (COMPOSE(prev, crt) == (CODE_IDENTITY) || COMPOSE(prev, crt) == reference) && crt_score!= MIN_SCORE && crt_score < 0) ? -2*crt_score : crt_score;
  //return (prev && (COMPOSE(prev, crt) == (CODE_IDENTITY) || COMPOSE(prev, crt) == reference) && crt_score!= MIN_SCORE && crt_score <= 0) ? 1 : crt_score;
}
inline ScoreType alignment__color_correction_gap_penalty_adjustment(CODE_TYPE prev, CODE_TYPE crt, ScoreType compensation_score) {
  return (prev && COMPOSE(prev, crt) == CODE_IDENTITY) ? compensation_score : 0;
}
/**
 * Suitable for affine gaps. According to the current alignment direction and the previous cell, establishes
 * if the gap is opened or extended.
 * @param alignment
 * @param prev_cell
 * @param direction The relative position of the current cell wrt prev_cell, encoded as I_MASK & vertical_dif | J_MASK & horizontal_dif.
 * @returns According to the case, either the gap_open or the gap_extension penalty, stored in the alignment's parameter field.
 */
inline ScoreType alignment__gap_penalty(AlignmentType * alignment, AlignmentCellType *prev_cell, DirectionType direction) {
  return alignment->params.gap[INDEX_FOR(prev_cell->from ^ direction)];
}
/**
 * Deduces the score of an aligned pair of colors, composed as color_match
 * @param alignment The address of the alignment data structure
 * @param color_match The composition of the two colors
 * @param quality_level The quality level of the read
 * @return The corresponding score
 */
inline ScoreType alignment__score_for_match(AlignmentType * alignment, CODE_TYPE color_match, short quality_level) {
  return alignment->params.pair_scores[INDEX_FOR(color_match)][quality_level] ;
}
/**
 * Deduces the score of an aligned pair of colors
 * @param alignment The address of the alignment data structure
 * @param color_1 The first color
 * @param color_2 The second color.
 * @param quality_level The quality level of the read
 * @return The corresponding score
 */
inline ScoreType alignment__score_for_pair(AlignmentType * alignment, CODE_TYPE color_1, CODE_TYPE color_2, short quality_level) {
  return alignment__score_for_match(alignment, COMPOSE(color_1, color_2), quality_level) ;
}
/**
 * Perform the actual gapped alignment.
 * @param alignment The address of the alignment data structure
 * @return 0 if we cannot align with less than the accepted number of mismatches in the first part of the read, 1 if we can
 */
inline int alignment__align_gapped_quick(AlignmentType * alignment, ScoreType min_accepted) {
  int i, j;
  AlignmentCellType *crt_cell, *top, *left, *corner;
  int mismatches_so_far;
  int significant_prefix_len = 3 * alignment->read_len / 4; /* TODO change this hard-coded value */
  alignment__init_martix(alignment);
  // Now starting at 1, since the first line should stay as it was initialized
  for (i = 1; i < significant_prefix_len; ++i) {
    mismatches_so_far = alignment->read_len;
    for (j = REF_START(i); j < REF_END(i); ++j) {
      crt_cell = alignment__matrix_access__no_check(alignment, i, j);
      /* Get the 3 cases that can contribute to the current one's score */
      corner = alignment__matrix_access__no_check(alignment, i-1, j-1);
      top = alignment__matrix_access__no_check(alignment, i-1, j);
      left = alignment__matrix_access__no_check(alignment, i, j-1);

      /* find out which one does contribute */
      crt_cell->mismatches  = corner->mismatches + (crt_cell->color_checksum && 1);
      if ((top->indels < alignment->params.allowed_indels) && (top->mismatches < crt_cell->mismatches)) {
       crt_cell->mismatches = top->mismatches;
       crt_cell->indels = top->indels + 1;
      }
      if ((left->indels < alignment->params.allowed_indels) && (left->mismatches < crt_cell->mismatches)) {
        crt_cell->mismatches = left->mismatches;
        crt_cell->indels = left->indels + 1;
      }
      if (mismatches_so_far > crt_cell->mismatches) {
        mismatches_so_far = crt_cell->mismatches;
      }
    }
    if (mismatches_so_far > DEFAULT_MAX_MISMATCHES) {
      return 0;
    }
  }
  return 1;
}

/**
 * Perform the actual gapped alignment.
 * @param alignment The address of the alignment data structure
 * @return The score of the best alignment
 */
//inline
int alignment__align_gapped(AlignmentType * alignment, ScoreType min_accepted) {
  int i, j;
  AlignmentCellType *crt_cell, *top, *left, *corner, *it;
  ScoreType tmp_score, best_score, max_on_line, match_score;
  ScoreType best_possible = (alignment->read_len - 2) * alignment->params.pair_scores[SYMBOL_PAIR_MATCH_IDX][MAX_READ_QUALITY_LEVEL];
  // initialize the alignment matrix
//  printf("%d %d %d %d\n", alignment->best_i, alignment->best_j, alignment->best_i0, alignment->best_j0);
//  return MIN_SCORE;
  alignment__init_martix(alignment);

  // Now starting at 1, since the first line should stay as it was initialized
  for (i = 1; i < alignment->read_len; ++i) {
    max_on_line = MIN_SCORE;
    for (j = REF_START(i); j < REF_END(i); ++j) {
      crt_cell = alignment__matrix_access__no_check(alignment, i, j);

      /* Get the 3 cases that can contribute to the current one's score */
      corner = alignment__matrix_access__no_check(alignment, i-1, j-1);
      top = alignment__matrix_access__no_check(alignment, i-1, j);
      left = alignment__matrix_access__no_check(alignment, i, j-1);

      /* find out which one does contribute */
      it = corner;
      match_score = alignment__color_correction_score_adjustment(
          corner->color_checksum, crt_cell->color_checksum, corner->color_autocorrect, crt_cell->score
      );
      crt_cell->score = corner->score + match_score;

      if ((top->indels < alignment->params.allowed_indels) && (tmp_score = top->score + alignment__gap_penalty(alignment, top, I_MASK)) > crt_cell->score) {
        it = top;
        crt_cell->score = tmp_score;
      }
      if ((left->indels < alignment->params.allowed_indels) && (tmp_score = left->score + alignment__gap_penalty(alignment, left, J_MASK)) > crt_cell->score) {
        it = left;
        crt_cell->score = tmp_score;
      }

      if (it == corner) {
        crt_cell->color_checksum = COMPOSE(crt_cell->color_checksum, it->color_checksum);
        crt_cell->indels = it->indels;
        // TODO optimize
        if (it->color_checksum != crt_cell->color_checksum) { /* mismatch in the crt cell*/
          if (it->color_checksum == CODE_IDENTITY) {
            /* can be the beginning of a SNP or can be a reading error */
            crt_cell->color_autocorrect = crt_cell->color_checksum;
            crt_cell->color_mismatch_freshness = COLOR_MISMATCH_PERSISTENCE(alignment->read_quality[i]);
          } else if (crt_cell->color_checksum == it->color_autocorrect) {
            /* ended a SNP */
            crt_cell->color_autocorrect = CODE_IDENTITY;
            crt_cell->color_checksum = CODE_IDENTITY;
            crt_cell->color_mismatch_freshness = COLOR_MISMATCH_NONE;
          } else if (crt_cell->color_checksum == CODE_IDENTITY) {
            if (it->color_mismatch_freshness - 1 > 0) {
              crt_cell->color_autocorrect = it->color_autocorrect;
              crt_cell->color_mismatch_freshness = it->color_mismatch_freshness - 1;
            } else {
              crt_cell->color_autocorrect = CODE_IDENTITY;
              crt_cell->color_mismatch_freshness = COLOR_MISMATCH_NONE;
            }
          } else { /**/
            if (it->color_mismatch_freshness - 1 > 0) {
              crt_cell->color_autocorrect = it->color_autocorrect;
              crt_cell->color_mismatch_freshness = it->color_mismatch_freshness - 1;
            } else {
              crt_cell->color_autocorrect = crt_cell->color_checksum;
              crt_cell->color_mismatch_freshness = COLOR_MISMATCH_PERSISTENCE(alignment->read_quality[i]);
            }
          }
        } else { /* match */
          if (it->color_autocorrect != CODE_IDENTITY) {
            /* 'deprecate' the previous mismatch */
            if (it->color_mismatch_freshness - 1 > 0) {
              if (crt_cell->color_checksum == CODE_IDENTITY) {
                crt_cell->color_autocorrect = CODE_IDENTITY;
                crt_cell->color_mismatch_freshness = COLOR_MISMATCH_NONE;
	      } else {
                crt_cell->color_autocorrect = it->color_autocorrect;
                crt_cell->color_mismatch_freshness = it->color_mismatch_freshness - 1;
              }
            }
	  } else {
            crt_cell->color_checksum = CODE_IDENTITY;
            crt_cell->color_autocorrect = CODE_IDENTITY;
            crt_cell->color_mismatch_freshness = COLOR_MISMATCH_NONE;
          }
        }
      } else if (it == top) {
        crt_cell->color_checksum = COMPOSE(top->color_checksum, alignment->read[i]);
        crt_cell->from = I_MASK;
        crt_cell->indels = it->indels + 1;
        if (it->color_mismatch_freshness - 1 > 0) {
          crt_cell->color_mismatch_freshness = it->color_mismatch_freshness - 1;
          crt_cell->color_autocorrect = corner->color_autocorrect;
        }
      } else { /* it == left */
        crt_cell->color_checksum = COMPOSE(left->color_checksum, alignment->reference[j]);
        crt_cell->from = J_MASK;
        crt_cell->indels = it->indels + 1;
        if (it->color_mismatch_freshness - 1 > 0) {
          crt_cell->color_mismatch_freshness = it->color_mismatch_freshness - 1;
          crt_cell->color_autocorrect = corner->color_autocorrect;
        }
      }
      if (max_on_line < crt_cell->score) {
        max_on_line = crt_cell->score;
      }
    }
    // estimate if the alignment has the chance to reach the minimum accepted quality;
    // if so, continue the alignment;
    // otherwise, abort and return MIN_SCORE;
    if (max_on_line + best_possible < min_accepted) {
      return MIN_SCORE;
    }
    best_possible -= alignment->params.pair_scores[SYMBOL_PAIR_MATCH_IDX][MAX_READ_QUALITY_LEVEL];
  }
  // find the maximum on the right bottom border, and retrieve it as the best alignment score
  best_score = MIN_SCORE;
  // bottom:
  i = alignment->read_len - 1;
  for (j = REF_START(i); j < REF_END(i); ++j) {
    if (best_score < alignment__get_score__no_check(alignment, i, j)) {
      best_score = alignment__get_score__no_check(alignment, i, j);
      alignment->best_i = i;
      alignment->best_j = j;
    }
  }
  --i;
  //right
  j = REF_END(i);
  while (j == alignment->ref_len) {
    --j;
    if (best_score < alignment__get_score__no_check(alignment, i, j)) {
      best_score = alignment__get_score__no_check(alignment, i, j);
      alignment->best_i = i;
      alignment->best_j = j;
    }
    --i;
    j = REF_END(i);
  }
  return best_score;
}
/**
 * Perform the actual alignment. First try ungapped, as it is faster; if it fails, try gapped.
 * @param alignment The address of the alignment data structure
 * @return The score of the best alignment
 */
int alignment__align(AlignmentType * alignment, ScoreType min_accepted) {
  alignment->status = GAPPED;
  //return alignment__align_gapped_quick(alignment, min_accepted) ? alignment__align_gapped(alignment, min_accepted) : MIN_SCORE;
  return alignment__align_gapped(alignment, min_accepted);
}

/**
 * Reconstructs the alignment from the matrix
 * @param alignment The address of the alignment data structure
 * @param i The line number of the best scoring cell (where the best alignment ends)
 * @param j The column number of the best scoring cell (where the best alignment ends)
 * @return A string containing the representation of this alignment (which the to_display field in the alignment structure)
 */
char* alignment__traceback(AlignmentType *alignment) {
  AlignmentCellType *cell;
  int i = alignment->best_i, j = alignment->best_j;
  size_t al_size = (size_t)(alignment->read_len + alignment->ref_len); //worse than worst case
  char *read = NULL, *ref = NULL, *m = NULL;
  int k=0, d, di, dj;
  int indel_idx = alignment->params.allowed_indels - 1;

  if (alignment->status == UNALIGNED || i < 0 || j < 0) {
    return NULL;
  }

  CODE_TYPE color = CODE_IDENTITY;

  alignment->traceback_seq_len = 0;

#ifdef TRY_UNGAPPED
  if (alignment->status == UNGAPPED) {
     if (VERBOSITY) {
       d = 0;
       for (k = 0; k < alignment->read_len; ++k) {
         alignment->to_display[d++] = COLOR_CODE[(int)(alignment->reference + alignment->params.allowed_indels)[k]];
       }
       alignment->to_display[d++] = '\n';
       for (k = 0; k < alignment->read_len; ++k) {
         alignment->to_display[d++] = (alignment->read[k] == (alignment->reference + alignment->params.allowed_indels)[k]) ? DISPLAYED_MATCH_SYMBOL : DISPLAYED_MISMATCH_SYMBOL;
       }
       alignment->to_display[d++] = '\n';
       for (k = 0; k < alignment->read_len; ++k) {
         alignment->to_display[d++] = COLOR_CODE[(int)alignment->read[k]];
       }
       alignment->to_display[d++] = '\n';
       alignment->to_display[d++] = '\0';
     }
     memset( alignment->read_indel_positions, 0x0, alignment->params.allowed_indels * sizeof(short));
     alignment->best_i0 = 0;
     alignment->best_j0 = alignment->params.allowed_indels;
     return alignment->to_display;
  }
#endif
  if (VERBOSITY) {
  // only build the displayed alignment if the verbosity level allows to display it
	  SAFE_FAILURE__ALLOC(read, al_size, char);
	  SAFE_FAILURE__ALLOC(ref, al_size, char);
	  SAFE_FAILURE__ALLOC(m, al_size, char);
  }
  while ((cell = alignment__matrix_access(alignment, i, j))) {
    alignment->best_i0 = i;
    alignment->best_j0 = j;
    if (VERBOSITY) {
      // only build the displayed alignment if the verbosity level allows to display it
      read[k] = (FROM_LEFT(cell->from)) ? DISPLAYED_SEQ_GAP_SYMBOL : CHARDISPLAY(alignment->read[i]);//( FROM_TOP(cell->from) ? CHARDISPLAY(alignment->read[i - 1])       : CHARDISPLAY(alignment->read[i]));
      ref [k] = ( FROM_TOP(cell->from)) ? DISPLAYED_SEQ_GAP_SYMBOL : CHARDISPLAY(alignment->reference [j]); //(FROM_LEFT(cell->from) ? CHARDISPLAY(alignment->reference [j - 1]) : CHARDISPLAY(alignment->reference [j]));
      m   [k] = (FROM_CORNER(cell->from)) ? ((alignment->read[i] == alignment->reference[j]) ? DISPLAYED_MATCH_SYMBOL : DISPLAYED_MISMATCH_SYMBOL) : DISPLAYED_MATCH_GAP_SYMBOL;
      ++k;
    }
    di = D_I(cell->from);
    dj = D_J(cell->from);
    /* if there was a gap, mark the position of the indel:
     *  0  => no more indels
     *  N  => insertion in the read at position N-1
     * -N  => deletion  in the read at position N-1 */
    if (di && !dj) {
      alignment->read_indel_positions[indel_idx--] = i + 1;
      TRACEBACK_SEQ_SET(alignment->traceback, alignment->traceback_seq_len, TRACEBACK_INSERT(alignment->read[i]));
    } else if (dj && !di) {
      alignment->read_indel_positions[indel_idx--] = -(i + 1);
      TRACEBACK_SEQ_SET(alignment->traceback, alignment->traceback_seq_len, TRACEBACK_DELETE(Ox0));
    } else {
      if (alignment->read[i] != alignment->reference[j]) {
        if (color || (i != 0 && j != 0 && alignment__matrix_access(alignment, i, j)->score > alignment__matrix_access(alignment, i-1, j-1)->score)) {
          TRACEBACK_SEQ_SET(alignment->traceback, alignment->traceback_seq_len, TRACEBACK_BASE_MISMATCH(alignment->read[i]));
          color = COMPOSE(color, COMPOSE(alignment->read[i], alignment->reference[j]));
        } else {
          TRACEBACK_SEQ_SET(alignment->traceback, alignment->traceback_seq_len, TRACEBACK_COLOR_MISMATCH(alignment->read[i]));
        }
      } else {
        TRACEBACK_SEQ_SET(alignment->traceback, alignment->traceback_seq_len, TRACEBACK_MATCH(0x0));
      }
    }
    ++(alignment->traceback_seq_len);
    i += di;
    j += dj;
  }

  if (VERBOSITY) {
  // only build the displayed alignment if the verbosity level allows to display it
    al_size = strlen(read) + strlen(ref) + strlen(m) + 4; /* 3 newlines and one \0 */
    d = 0;
    for (k = strlen(ref) - 1; k >= 0; --k) {
      alignment->to_display[d++] = ref[k];
    }
    alignment->to_display[d++] = '\n';
    for (k = strlen(m) - 1; k >= 0; --k) {
      alignment->to_display[d++] = m[k];
    }
    alignment->to_display[d++] = '\n';
    for (k = strlen(read) - 1; k >= 0; --k) {
      alignment->to_display[d++] = read[k];
    }
    alignment->to_display[d++] = '\n';
    alignment->to_display[d++] = '\0'; // otherwise some leftovers from previous, longer alignments may show up...

    free(read);
    free(ref);
    free(m);
  }
  // shift the indel positions if necessary
  d = -1;
  while (alignment->read_indel_positions[++d] == 0);
  if (d) {
    for (indel_idx = 0; indel_idx + d < alignment->params.allowed_indels; ++indel_idx) {
      alignment->read_indel_positions[indel_idx] = alignment->read_indel_positions[indel_idx + d];
      alignment->read_indel_positions[indel_idx + d] = 0;
    }
  }
  return alignment->to_display;
}

/**
 * Checks if an alignment score is good enough
 * @param score
 * @return 1 if the score is above an established threshold, 0 otherwise
 */
int alignment__score_accepted(int score) {
  // TODO re-implement this with a better condition
  return (score >= min_accepted_score);
}

/**
 * Access the alignment matrix, with coordinates that are not "aware" of the matrix' sparsness
 * @param alignment The address of the alignment data structure
 * @param i The requested cell's line coordinate
 * @param j The requested cell's column coordinate, as a number from 0 to reference length (to be transformed in actual matrix coords)
 * @return A reference to the requested matrix cell, or NULL if the parameters are out of range
 */
inline AlignmentCellType* alignment__matrix_access(AlignmentType * alignment, int i, int j) {
  return (CHECK_RELATIVE_COORDS(i, j)) ? alignment->matrix[i][TO_MATRX_ABS_COORD(j, i)] : NULL;
}
/**
 * Access the alignment matrix, with coordinates that are not "aware" of the matrix' sparsness, without checking the parameter range
 * @param alignment The address of the alignment data structure
 * @param i The requested cell's line coordinate
 * @param j The requested cell's column coordinate, as a number from 0 to reference length (to be transformed in actual matrix coords)
 * @return A reference to the requested matrix cell,
 */
inline AlignmentCellType* alignment__matrix_access__no_check(AlignmentType * alignment, int i, int j) {
  return alignment->matrix[i][TO_MATRX_ABS_COORD(j, i)];
}

/**
 * Access the alignment matrix, with the true matrix coordinates
 * @param alignment The address of the alignment data structure
 * @param i The requested cell's line coordinate
 * @param j The requested cell's column coordinate, as a number from 0 to matrix line length
 * @return A reference to the requested matrix cell, or NULL if the parameters are out of range
 */
inline AlignmentCellType* alignment__matrix_absolute_access(AlignmentType * alignment, int i, int j) {
  return (CHECK_ABSOLUTE_COORDS(i, j)) ? alignment->matrix[i][j] : NULL;
}
/**
 * Access the alignment matrix, with the true matrix coordinates, without checking the parameter range
 * @param alignment The address of the alignment data structure
 * @param i The requested cell's line coordinate
 * @param j The requested cell's column coordinate, as a number from 0 to matrix line length
 * @return A reference to the requested matrix cell, or NULL if the parameters are out of range
 */
inline AlignmentCellType* alignment__matrix_absolute_access__no_check(AlignmentType * alignment, int i, int j) {
  return alignment->matrix[i][j];
}
/**
 * Retrieve a score in the alignment matrix, with coordinates that are not "aware" of the matrix' sparseness
 * @param alignment The address of the alignment data structure
 * @param i The requested cell's line coordinate
 * @param j The requested cell's column coordinate, as a number from 0 to reference length (to be transformed in actual matrix coords)
 * @return The found score, or MIN_SCORE if the parameters are out of range
 */
inline ScoreType alignment__get_score(const AlignmentType * alignment, int i, int j) {
  return (CHECK_RELATIVE_COORDS(i, j)) ? alignment->matrix[i][TO_MATRX_ABS_COORD(j, i)]->score : MIN_SCORE;
}
/**
 * Retrieve a score in the alignment matrix, with coordinates that are not "aware" of the matrix' sparseness, without checking the parameter range
 * @param alignment The address of the alignment data structure
 * @param i The requested cell's line coordinate
 * @param j The requested cell's column coordinate, as a number from 0 to reference length (to be transformed in actual matrix coords)
 * @return The found score, or MIN_SCORE if the parameters are out of range
 */
inline ScoreType alignment__get_score__no_check(const AlignmentType * alignment, int i, int j) {
  return alignment->matrix[i][TO_MATRX_ABS_COORD(j, i)]->score;
}

/**
 * Retrieve a score in the alignment matrix, with with the true matrix coordinates
 * @param alignment The address of the alignment data structure
 * @param i The requested cell's line coordinate
 * @param j The requested cell's column coordinate, as a number from 0 to matrix line length
 * @return The found score, or MIN_SCORE if the parameters are out of range
 */
inline ScoreType alignment__get_score_absolute(const AlignmentType * alignment, int i, int j) {
  return (CHECK_ABSOLUTE_COORDS(i, j)) ? alignment->matrix[i][j]->score : MIN_SCORE;
}
/**
 * Retrieve a score in the alignment matrix, with with the true matrix coordinates, without checking the parameter range
 * @param alignment The address of the alignment data structure
 * @param i The requested cell's line coordinate
 * @param j The requested cell's column coordinate, as a number from 0 to matrix line length
 * @return The found score, or MIN_SCORE if the parameters are out of range
 */
inline ScoreType alignment__get_score_absolute__no_check(const AlignmentType * alignment, int i, int j) {
  return alignment->matrix[i][j]->score;
}
