#include "genome_map.h"

#define GENOME_MAP_SIZE (genome_map->ref_db->size + genome_map->hitmap->indel_count)
#define GENOME_MAP_READS_COUNT genome_map->hitmap->size
#define GENOME_MAP_INDELS genome_map->hitmap->indel_count

#define TRUSTED_COLOR_COUNT_RELATIVE_DIFFERENCE 0.6
#define FORWARD_WINDOW_SIZE 4

#define GENOME_MAP_NO_COLOR COLOR_CODE_COUNT


#define COLOR_ORDER__EXISTING_COLOR_FLAG 0x8
#define COLOR_ORDER__EXISTING_COLOR_MASK 0x7
#define COLOR_ORDER__BITS_PER_COLOR 4
#define COLOR_ORDER__REAL_POS(p) (p << 2)
#define COLOR_ORDER__ADD(i, value, pos) {i |= (((value) | COLOR_ORDER__EXISTING_COLOR_FLAG) << COLOR_ORDER__REAL_POS(pos));}
#define COLOR_ORDER__POP(i, result)  {result = (i) & COLOR_ORDER__EXISTING_COLOR_MASK; i >>= COLOR_ORDER__BITS_PER_COLOR;}
#define COLOR_ORDER__HAS_MORE(i)  ((i) & COLOR_ORDER__EXISTING_COLOR_FLAG)

#ifdef COMPRESSED_COLOR_COUNTERS

#define COLOR_COUNTER_TYPE unsigned

#define COLOR_COUNTER_MASK 0x3F
#define COLOR_COUNTER_MASK_SIZE 6

#define COLOR_COUNTER_MASK_SHIFT(color) ((color) * COLOR_COUNTER_MASK_SIZE)
#define COLOR_COUNTER_MASK_FOR(color) (COLOR_COUNTER_MASK << COLOR_COUNTER_MASK_SHIFT(color))
#define COLOR_COUNTER_GET(counter, color) (((counter) >> COLOR_COUNTER_MASK_SHIFT(color)) & COLOR_COUNTER_MASK)
#define COLOR_COUNTER_SET(counter, color, value) counter = (counter & (~ COLOR_COUNTER_MASK_FOR(color))) | (((value) & COLOR_COUNTER_MASK) << COLOR_COUNTER_MASK_SHIFT(color));
#define COLOR_COUNTER_INC(counter, color) COLOR_COUNTER_SET(counter, color, (COLOR_COUNTER_GET(counter, color) + 1));


//#define COLOR_COUNTER_GET(counter, color) counter
//#define COLOR_COUNTER_SET(counter, color, value) counter = value;
//#define COLOR_COUNTER_INC(counter, color) ++counter;

#else

#define COLOR_COUNTER_TYPE int*

#define COLOR_COUNTER_GET(counter, color) counter[color]
#define COLOR_COUNTER_SET(counter, color, value) { counter[color] = value ; }
#define COLOR_COUNTER_INC(counter, color) { counter[color]++; }

#endif

extern char** cmd_line;
extern int    cmd_line_len;

int map_greedy = 0;


/* #define FILTER_AMBIGUOUS_READS */

/**
 * Access to map positions: get
 */
inline int genome_map__get_read_idx_at(const GenomeMapType* genome_map, const int position) {
	return  genome_map->g_map[position] ? genome_map->g_map[position]->read_idx : UNMAPPED;
}
inline int genome_map__get_read_score_rank_at(const GenomeMapType* genome_map, const int position) {
	return  genome_map->g_map[position] ? genome_map->g_map[position]->score_rank : UNMAPPED;
}
/**
 * Create the data structure for the needed (mapped) positions
 */
inline void genome_map__create_position_data(GenomeMapType* genome_map, int position) {
  int k;
  if (genome_map->g_map[position]) {
    return;
  }
  SAFE_FAILURE__ALLOC(genome_map->g_map[position], 1, GenomeMapPositionType);
  genome_map->g_map[position]->read_idx = UNMAPPED;
  if (!map_greedy) {
#ifdef COMPRESSED_COLOR_COUNTERS
    genome_map->g_map[position]->aligned_positions_freqs = 0;
    if (GENOME_MAP_INDELS) {
      SAFE_FAILURE__ALLOC(genome_map->g_map[position]->inserted_positions_freqs, genome_map->hitmap->indel_count, unsigned int);
      for (k = 0; k < GENOME_MAP_INDELS; ++k) {
         genome_map->g_map[position]->inserted_positions_freqs[k] = 0;
      }
    }
#else
    int j;
    SAFE_FAILURE__ALLOC(genome_map->g_map[position]->aligned_positions_freqs, (COLOR_CODE_COUNT + 1), int)
    if (GENOME_MAP_INDELS) {
      SAFE_FAILURE__ALLOC(genome_map->g_map[position]->inserted_positions_freqs, genome_map->hitmap->indel_count, int*);
      for (j = 0; j < genome_map->hitmap->indel_count ; ++j) {
        SAFE_FAILURE__ALLOC(genome_map->g_map[position]->inserted_positions_freqs[j], COLOR_CODE_COUNT + 1, int);
      }
    }
    for (j = 0; j < COLOR_CODE_COUNT + 1; ++j) {
      genome_map->g_map[position]->aligned_positions_freqs[j] = 0;
    }
    genome_map->g_map[position]->aligned_positions_freqs[NTH_CODE(genome_map->ref_db->sequence, position)] = 1;
    if (genome_map->hitmap && genome_map->g_map[position]->inserted_positions_freqs) {
      for (k = 0; k < GENOME_MAP_INDELS; ++k) {
        for (j = 0; j < COLOR_CODE_COUNT + 1; ++j) {
          genome_map->g_map[position]->inserted_positions_freqs[k][j] = 0;
        }
      }
    }
#endif
  }
}
/**
 * Access to map positions: set
 */
inline void genome_map__set_read_idx_at(GenomeMapType* genome_map, int position, int ri, int score_rank) {
	if (!genome_map->g_map[position]) {
		genome_map__create_position_data(genome_map, position);
	} else if (genome_map->g_map[position]->read_idx >= 0 && genome_map->g_map[position]->read_idx < genome_map->reads_db->size) {
		 genome_map->reads_db->reads[genome_map->g_map[position]->read_idx].disabled = 0;
	}
	genome_map->g_map[position]->read_idx = ri;
	genome_map->g_map[position]->score_rank = score_rank;
}
/**
 * Creates an empty genome map
 * @param genome_map The map to create
 * @param hitmap The hitmap that will help populate the genome map
 * @param ref_db The reference sequence
 */
GenomeMapType* genome_map__create(HitMapType* hitmap, ReferenceDBType* ref_db, ReadsDBType* reads_db) {
	int i;
	GenomeMapType* genome_map;
	SAFE_FAILURE__ALLOC(genome_map, 1, GenomeMapType);
	genome_map->ref_db   = ref_db;
	genome_map->reads_db = reads_db;
	genome_map->hitmap = hitmap;
	SAFE_FAILURE__ALLOC(genome_map->g_map, GENOME_MAP_SIZE, GenomeMapPositionType*);
	SAFE_FAILURE__ALLOC(genome_map->sorted_read_indices, GENOME_MAP_READS_COUNT, int);
    for (i = 0; i < GENOME_MAP_READS_COUNT; ++i) {
        genome_map->sorted_read_indices[i] = i;
    }
	genome_map->hitmap = hitmap;
	return genome_map;
}

/* ********************************************************************************
 *  Mapping version 1: greedy
 */
/**
 * Attempt to map a read
 */
static int genome_map__map_read_greedy(GenomeMapType* genome_map, int read_id, int start_at) {
  int i = start_at, where, other_read_id, rank1, rank2 = 0;
  while (i < MAP_DETAIL_SIZE && (where = genome_map->hitmap->map[read_id][i].ref_position) != UNMAPPED) {
    other_read_id = genome_map__get_read_idx_at(genome_map, where);
    if (other_read_id == UNMAPPED) {
      genome_map__set_read_idx_at(genome_map, where, read_id, i);
      return where;
    } else if (hit_map__score_for(genome_map->hitmap, read_id, where, &rank1) > hit_map__score_for(genome_map->hitmap, other_read_id, where, &rank1)) {
      genome_map__map_read_greedy(genome_map, other_read_id, rank2+1);
      genome_map__set_read_idx_at(genome_map, where, read_id, i);
      return where;
    }
    ++i;
  }
  return UNMAPPED;
}
/**
 * Build a genome map for a database of mapped reads - greedy approach
 */
void genome_map__build_greedy(GenomeMapType* genome_map) {
	int i, coverage = 0;
	for (i = 0; i < genome_map->hitmap->size; ++i) {
		genome_map__map_read_greedy(genome_map, i, 0);
	}
	for (i = 0; i < genome_map->ref_db->size; ++i) {
	  if (genome_map__get_read_idx_at(genome_map, i) != UNMAPPED) {
	    genome_map->reads_db->reads[genome_map__get_read_idx_at(genome_map, i)].disabled = 1;
	    ++coverage;
	  }
	}
	if (VERBOSITY >= VERBOSITY_MODERATE) {
	  printf("Reference coverage: %5.2f%%. Actually mapped reads: %5.2f%% (%d).\n\nGenerating output...\n", coverage * 100.0 / genome_map->ref_db->size, coverage * 100.0 / genome_map->reads_db->size, coverage);
	}
}

/* ********************************************************************************
 *  Mapping version 2: read multiple alignment based
 */
static inline ScoreType genome_map__get_sorted_score(GenomeMapType* genome_map, int index) {
    return genome_map->hitmap->map[genome_map->sorted_read_indices[index]][0].score;
}

static inline HitInfoType* genome_map__get_sorted_hit(GenomeMapType* genome_map, int index) {
    return genome_map->hitmap->map[genome_map->sorted_read_indices[index]];
}

inline void genome_map__radix_sort_reads(GenomeMapType* genome_map) {
    int i = 0, j;
    int N = genome_map->hitmap->size;
#define BITS_IN_DIGIT 4
#define BUCKETS 16 /* 1 << BITS_IN_DIGIT */
#define DIGITS 4   /* sizeof(ScoreType) * 8 / BITS_IN_DIGIT */
#define BUCKET_MASK 0xF   /* BUCKETS - 1 */
#define GET_BUCKET(i, j) (BUCKET_MASK - ((genome_map__get_sorted_score(genome_map, (j)) >> ((i) * BITS_IN_DIGIT) ) & BUCKET_MASK))
    int count[BUCKETS];
    int index[BUCKETS];
    int *swap;
    int *tmp;
	SAFE_FAILURE__ALLOC(tmp, N, int);

    while (i < DIGITS) {
        /* Count how many entries are in each bucket */
        memset (count, 0, sizeof (count));
	    for (j = 0; j < N; ++j) {
	        ++ count[GET_BUCKET(i, j)];
	    }
        /* Determine the position of each bucket in the destination */
	    index[0] = 0;
        for (j = 1; j < BUCKETS; ++j) {
            index[j] = index[j-1] + count[j-1];
        }
        /* Split into buckets */
        for (j = 0; j < N; ++j) {
	        tmp[index[GET_BUCKET(i, j)]++] = genome_map->sorted_read_indices[j];
	    }
	    /* Swap the two index vectors */
	    ++i;
	    swap = tmp;
	    tmp = genome_map->sorted_read_indices;
	    genome_map->sorted_read_indices = swap;
	}

    free(tmp);
}

/**
 * Find the color that appears the most often in the reads at the certain position.
 * If there is a tie, and the reference color is one of the most frequent colors,
 * then the reference color is returned.
 * Otherwise, the first color with the highest number of appearances is returned.
 */
inline int genome_map__get_direct_consensus_color(GenomeMapType* genome_map, int position, int offset, int reference_color) {
  int i, best_color = reference_color;
  COLOR_COUNTER_TYPE counts;
  if (genome_map->g_map[position] == NULL) {
    return reference_color;
  }
  counts = (offset == -1) ? genome_map->g_map[position]->aligned_positions_freqs :  genome_map->g_map[position]->inserted_positions_freqs[offset];
  for (i = 0; i <= COLOR_CODE_COUNT; ++i) {
	/*
	 if (counts[i] > counts[best_color]) {
      best_color = i;
     }
	 */
	if (COLOR_COUNTER_GET(counts, i) > COLOR_COUNTER_GET(counts, best_color)) {
	   best_color = i;
	}
  }
  /*
  if (counts[best_color] == 0) {
    best_color = GENOME_MAP_NO_COLOR;
  }
  */
  if (COLOR_COUNTER_GET(counts, best_color) == 0) {
    best_color = GENOME_MAP_NO_COLOR;
  }
  return best_color;
}

/**
 * Find the colors that appears the most often in the reads at the certain position.
 * The result is an integer, holding the IDs of each color, in decreasing order of their
 * frequency of appearance on that position, encoded on 4 bits starting with the least
 * significant. Only the colors which appear a significant number of times are considered.
 * (the condition is to appear at least SIGNIFICANT_PERCENTAGE % wrt the previous color).
 * The 4 bits allocated to each color should be interpreted as follows:
 * FCCC, where F is a flag, set to 1 if those 4 bits correspond to a color,
 * and CCC is the 3 bit representation of the 5 possible values, from 0 to 4.
 */
inline int genome_map__get_direct_consensus_colors(int* counts, int reference_color) {
  int i, result = 0;
  if (counts == NULL) {
	  return reference_color;
  }
#define ACCEPTANCE_THRESHOLD 0.6
  static int result_indices[COLOR_CODE_COUNT + 1];
  sort__indices(counts, result_indices, COLOR_CODE_COUNT + 1);

  if (counts[result_indices[0]] == 0) {
    COLOR_ORDER__ADD(result, GENOME_MAP_NO_COLOR, 0);
  } else {
    COLOR_ORDER__ADD(result, result_indices[0], 0);
    for (i = 1; i <= COLOR_CODE_COUNT; ++i) {
      if (counts[result_indices[i]] >= counts[result_indices[i-1]] * ACCEPTANCE_THRESHOLD) {
        COLOR_ORDER__ADD(result, result_indices[i], i);
      } else {
        break;
      }
    }
  }
  return result;
}

/**
 * Find the color that appears the most often in the reads at the certain position.
 * If there is a tie, and the reference color is one of the most frequent colors,
 * then the reference color is returned.
 * Otherwise, the first color with the highest number of appearances is returned.
 */
inline int genome_map__get_consensus_color(GenomeMapType* genome_map, int position, int offset, int reference_color) {
  static CODE_TYPE crt_checksum = CODE_IDENTITY;
  int best_color = reference_color;
  int window = 0;
  int fwd_i = position, fwd_j = offset;
  CODE_TYPE checksum = crt_checksum;
  best_color = genome_map__get_direct_consensus_color(genome_map, position, offset, reference_color);

  if (best_color != reference_color) {
      if (fwd_j == -1) {
        checksum = COMPOSE_SAFE(
            checksum,
            COMPOSE_SAFE(
              reference_color,
              best_color));
      } else {
        checksum = COMPOSE_SAFE(
          checksum,
          best_color);
      }
    while (window < FORWARD_WINDOW_SIZE) {
      ++fwd_j;
      if (fwd_j >= genome_map->hitmap->indel_count ||
        genome_map__get_direct_consensus_color(genome_map, fwd_i, fwd_j, GENOME_MAP_NO_COLOR) == GENOME_MAP_NO_COLOR) {
        fwd_j = -1;
        ++fwd_i;
      }
      if (fwd_i >= genome_map->ref_db->size) {
    	  break;
      }
      if (fwd_j == -1) {
        checksum = COMPOSE_SAFE(
            checksum,
            COMPOSE_SAFE(
              NTH_CODE(genome_map->ref_db->sequence, fwd_i),
              genome_map__get_direct_consensus_color(
                genome_map, fwd_i, -1, NTH_CODE(genome_map->ref_db->sequence, fwd_i))));
      } else {
        checksum = COMPOSE_SAFE(
          checksum,
          genome_map__get_direct_consensus_color(genome_map, fwd_i, fwd_j, GENOME_MAP_NO_COLOR));
      }
      if(checksum == CODE_IDENTITY) {
    	  break;
      }
      ++window;
    }
    if (checksum != CODE_IDENTITY) {
      if (offset == -1) {
        best_color = reference_color;
      } else {
        best_color = GENOME_MAP_NO_COLOR;
      }
    }
  }
  if (offset >= 0) {
      crt_checksum = COMPOSE_SAFE(crt_checksum, best_color);
  } else {
  	  crt_checksum = COMPOSE_SAFE(crt_checksum, COMPOSE_SAFE(
  	      reference_color, best_color
  	      ));
  }
  return best_color;
}

/**
 * Find the color that appears the most often in the reads at the certain position.
 * If there is a tie, and the reference color is one of the most frequent colors,
 * then the reference color is returned.
 * Otherwise, the first color with the highest number of appearances is returned.

inline int genome_map__get_consensus_color_full(GenomeMapType* genome_map, int position, int offset, int reference_color) {
  static CODE_TYPE crt_checksum = CODE_IDENTITY;
  int best_color;
  int window = 0;
  int *counts = (offset == -1) ? genome_map->g_map[position].aligned_positions_freqs : genome_map->g_map[position].inserted_positions_freqs[offset];
  int fwd_i = position, fwd_j = offset;
  CODE_TYPE checksum = crt_checksum;
  colors_order = genome_map__get_direct_consensus_colors(counts, reference_color);

  if (best_color != reference_color) {
      if (fwd_j == -1) {
        checksum = COMPOSE_SAFE(
            checksum,
            COMPOSE_SAFE(
              reference_color,
              best_color));
      } else {
        checksum = COMPOSE_SAFE(
          checksum,
          best_color);
      }
    while (window < FORWARD_WINDOW_SIZE) {
      ++fwd_j;
      if (fwd_j >= genome_map->hitmap->indel_count ||
          genome_map__get_direct_consensus_colors(
            genome_map->g_map[fwd_i].inserted_positions_freqs[fwd_j], GENOME_MAP_NO_COLOR) == GENOME_MAP_NO_COLOR) {
        fwd_j = -1;
        ++fwd_i;
      }
      if (fwd_i >= genome_map->ref_db->size) {
    	  break;
      }
      if (fwd_j == -1) {
        checksum = COMPOSE_SAFE(
            checksum,
            COMPOSE_SAFE(
              NTH_CODE(genome_map->ref_db->sequence, fwd_i),
              genome_map__get_direct_consensus_colors(
                genome_map->g_map[fwd_i].aligned_positions_freqs,
                NTH_CODE(genome_map->ref_db->sequence, fwd_i))));
      } else {
        checksum = COMPOSE_SAFE(
          checksum,
          genome_map__get_direct_consensus_color(
            genome_map->g_map[fwd_i].inserted_positions_freqs[fwd_j], GENOME_MAP_NO_COLOR));
      }
      if(checksum == CODE_IDENTITY) {
    	  break;
      }
      ++window;
    }
    if (checksum != CODE_IDENTITY) {
      if (offset == -1) {
        best_color = reference_color;
      } else {
        best_color = GENOME_MAP_NO_COLOR;
      }
    }
  }
  if (offset >= 0) {
      crt_checksum = COMPOSE_SAFE(crt_checksum, best_color);
  } else {
  	  crt_checksum = COMPOSE_SAFE(crt_checksum, COMPOSE_SAFE(
  	      reference_color, best_color
  	      ));
  }
  return best_color;
}
 */
/**
 * Update color frequency in stacked reads
 */
inline void genome_map__update_consensus_color(GenomeMapType* genome_map, int position, int offset, int color) {
  /* Only creates the data if null. This ensures counts != null */
  genome_map__create_position_data(genome_map, position);
  COLOR_COUNTER_TYPE counts = (offset == -1) ? genome_map->g_map[position]->aligned_positions_freqs : genome_map->g_map[position]->inserted_positions_freqs[offset];
  COLOR_COUNTER_INC(counts, color);
}
/**
 * Recompute the alignment score of a read, according to other reads it overlaps in the mapping
 */
ScoreType genome_map__compute_adjusted_score(GenomeMapType* genome_map, int ri, int rank) {
	ScoreType score = 0;
	int i, ref_position = genome_map->hitmap->map[ri][rank].ref_position, indel = 0;
	CODE_TYPE ref_color = 0, read_color = 0;
	unsigned char traceback_cell, traceback_cell_type;
	int traceback_len = genome_map->hitmap->map[ri][rank].traceback_seq_len;
	unsigned char* traceback = genome_map->hitmap->map[ri][rank].traceback;

	if (ref_position < 0) return MIN_SCORE; /* genome_map->hitmap->map[ri][rank].score */;

	for (i = 0; i < traceback_len; ++i) {
		traceback_cell = TRACEBACK_SEQ_GET(traceback, i);
		traceback_cell_type = GET_PAIR_TYPE(traceback_cell);
		if (traceback_cell_type == TRACEBACK_PAIR_MATCH){
			ref_color = NTH_CODE(genome_map->ref_db->sequence, ref_position);
			score += (ref_color ==
				genome_map__get_direct_consensus_color(genome_map, ref_position, -1, ref_color)) ? MATCH : MISMATCH;
			++ref_position;
		} else if(traceback_cell_type == TRACEBACK_PAIR_READ_DELETION){
			score += (genome_map__get_direct_consensus_color(
					genome_map, ref_position, indel++, GENOME_MAP_NO_COLOR)
					== GENOME_MAP_NO_COLOR) ? MATCH : MISMATCH;
			++ref_position;
		} else if (traceback_cell_type == TRACEBACK_PAIR_COLOR_MISMATCH || traceback_cell_type == TRACEBACK_PAIR_BASE_MISMATCH) {
			read_color = (traceback_cell & TRACEBACK_COLOR_MASK );
			score += (read_color ==
				genome_map__get_direct_consensus_color(genome_map, ref_position, -1, read_color)) ? MATCH : MISMATCH;
			++ref_position;
		} else if(traceback_cell_type == TRACEBACK_PAIR_READ_INSERTION){
			read_color = (traceback_cell & TRACEBACK_COLOR_MASK );
			score += (read_color ==
				genome_map__get_direct_consensus_color(genome_map, ref_position, indel++, read_color)) ? MATCH : MISMATCH;
		}
	}
	return score;
}

#ifdef FILTER_AMBIGUOUS_READS
static void filter_mapped_reads(GenomeMapType* genome_map) {
	int ri;
	for (ri = 0; ri < GENOME_MAP_READS_COUNT; ++ri) {
		if(genome_map->hitmap->map[ri][0].score <= min_trusted_score || genome_map->hitmap->map[ri][1].score >= min_trusted_score + MIN_TRUSTED_SCORE_DIF) {
			genome_map->reads_db->reads[ri].disabled = 1;
		}
	}
}
#endif

/**
 * Build a genome map for a database of mapped reads: multiple alignment approach
 */
void genome_map__build_contextual(GenomeMapType* genome_map) {
	int i, j, coverage = 0, ri, best_index, ref_position, insertions, p;
	ScoreType best_score, score;
	unsigned char traceback_cell, traceback_cell_type;
#ifdef FILTER_AMBIGUOUS_READS
	filter_mapped_reads(genome_map);
#endif
	// 1. sort reads by best score
	genome_map__radix_sort_reads(genome_map);
	// 2. for each read,
	for (i = 0; i < GENOME_MAP_READS_COUNT; ++i) {
		// get the target read
		ri = genome_map->sorted_read_indices[i];
		// skip reads already mapped on other references
		if (genome_map->reads_db->reads[ri].disabled) {
			continue;
		}
		best_index = UNMAPPED;
		best_score = MIN_SCORE;
		// 2.1  recompute the read scores wrt the "consensus"
		for (j = 0; j < MAP_DETAIL_SIZE && genome_map->hitmap->map[ri][j].ref_position != UNMAPPED; ++j) {
			score = genome_map->hitmap->map[ri][j].adjusted_score = genome_map__compute_adjusted_score(genome_map, ri, j);
			if (score > best_score &&
					genome_map__get_read_idx_at(genome_map, genome_map->hitmap->map[ri][j].ref_position) == UNMAPPED) {
				best_index = j;
				best_score = score;
			}
		}
		// 2.2  map the read in the best available place and update the consensus
		if (best_index != UNMAPPED) {
			// reference position covered
			++coverage;
			// disable read to prevent from mapping on other references
			genome_map->reads_db->reads[ri].disabled = 1;
			// map
			ref_position = genome_map->hitmap->map[ri][best_index].ref_position;
			genome_map__set_read_idx_at(genome_map, ref_position, ri, best_index);
			// adjust consensus
			insertions = -1;
			for (j = 0; j < genome_map->hitmap->map[ri][best_index].traceback_seq_len; ++j) {
				traceback_cell = TRACEBACK_SEQ_GET(genome_map->hitmap->map[ri][best_index].traceback, j);
			    traceback_cell_type = GET_PAIR_TYPE(traceback_cell);
			    if(traceback_cell_type == TRACEBACK_PAIR_READ_INSERTION){
			    	genome_map__update_consensus_color(genome_map, ref_position-1,  ++insertions, (traceback_cell & TRACEBACK_COLOR_MASK) /* read_color */);
			    } else {
					if (traceback_cell_type == TRACEBACK_PAIR_MATCH || traceback_cell_type == TRACEBACK_PAIR_COLOR_MISMATCH){
						genome_map__update_consensus_color(genome_map, ref_position, -1, NTH_CODE(genome_map->ref_db->sequence, ref_position) /* ref_color */);
					} else if(traceback_cell_type == TRACEBACK_PAIR_READ_DELETION){
						genome_map__update_consensus_color(genome_map, ref_position, -1, GENOME_MAP_NO_COLOR);
					} else if (/* traceback_cell_type == TRACEBACK_PAIR_COLOR_MISMATCH || */ traceback_cell_type == TRACEBACK_PAIR_BASE_MISMATCH) {
						genome_map__update_consensus_color(genome_map, ref_position, -1, (traceback_cell & TRACEBACK_COLOR_MASK ) /* read_color */);
					}
					if (ref_position && insertions < 0) {
					  for (p = 0; p < GENOME_MAP_INDELS; ++p) {
						genome_map__update_consensus_color(genome_map, ref_position-1, p, GENOME_MAP_NO_COLOR);
					  }
					}
			    	++ref_position; /* advance in the reference sequence */
			    	insertions = -1; /* reset insertion sequence */
			    }
			}
		}
	}
	if (VERBOSITY >= VERBOSITY_MODERATE) {
	   printf("Reference coverage: %5.2f%%. Actually mapped reads: %5.2f%% (%d).\n\nGenerating output...\n", coverage * 100.0 / genome_map->ref_db->size, coverage * 100.0 / genome_map->reads_db->size, coverage);
	}
}

void genome_map__build(GenomeMapType* genome_map) {
	if (map_greedy) {
		genome_map__build_greedy(genome_map);
	} else {
		genome_map__build_contextual(genome_map);
	}
}

/* *************************************************************** */
inline void genome_map__display_read_pos__colors(CODE_TYPE* read_sequence, int* position,
          short* indel_positions, int indel_positions_count, int* next_indel,
          //int* aligned_color_counts, int** inserted_color_counts,
	      GenomeMapPositionType* map_position,
          char* output_buffer) {
  char c;
  int i;
  CODE_TYPE color;
  if (!read_sequence) {
	// nothing mapped here; increment position and exit
    (*position)++;
    return;
  }
  if (*position < 0) {
    // haven't begun displaying yet
    c = (*position == -1) ? '@' : ' '; //read start marker
    (*position)++;
  } else if (*next_indel < indel_positions_count &&
      indel_positions[*next_indel] < 0 &&
      indel_positions[*next_indel] == -(*position) ){
    // deletion in the read at the given position
    c = SEQ_GAP_SYMBOL;
    // update the next indel's index
    (*next_indel)++;
    // gap symbol displayed, increase its count
#if(MAPPING_ALGO_SWITCHER == MAPPING_ALGO_GREEDY)
    map_position->aligned_positions_freqs[GENOME_MAP_NO_COLOR] ++;
#endif
  } else {
    // compute and set the next char
    color = NTH_CODE(read_sequence, (*position));
    c = COLOR_CODE[(int)color];
    (*position)++;

#if(MAPPING_ALGO_SWITCHER == MAPPING_ALGO_GREEDY)
    map_position->aligned_positions_freqs[(int)color] ++;
#endif
  }
  // print the char
  //fprintf(output, " %c", c);
  output_buffer[1] = c;
  i = 0;
  if (*position >= 0 && *next_indel < indel_positions_count && indel_positions[*next_indel]) {
    // there are indels left
	// insertions at position p must be displayed after position p-1
    while (*next_indel < indel_positions_count && indel_positions[*next_indel] == (*position) + 1) {
      // insertion in the read
      color = NTH_CODE(read_sequence, (*position));
      ++(*position);
      //fprintf(output, "%d", color);
      output_buffer[i + 2] = COLOR_CODE[(int)color];
      // update the next indel's index
      (*next_indel)++;
#if(MAPPING_ALGO_SWITCHER == MAPPING_ALGO_GREEDY)
      map_position->inserted_positions_freqs[i][(int)color]++;
#endif
      ++i;
    }
  }
#if(MAPPING_ALGO_SWITCHER == MAPPING_ALGO_GREEDY)
  // print spaces
  while (i != indel_positions_count) {
	  map_position->inserted_positions_freqs[i++][GENOME_MAP_NO_COLOR]++;
  }
#endif
}
//
//
#define BLOCK_SIZE 70
#define GET_AND_PRINT_BASE(fout, base, crt_color)\
{\
	base = TRANSFORM(base, crt_color);\
	fprintf(fout, "%c", BASE[(int)base]);\
	if (++result_len % BLOCK_SIZE == 0) {\
	   fprintf(fout, "\n");\
	}\


/**
 * Outputs the mapped reads
 * @param map
 * @param ref_db
 * @param reads_db
 * @param output
 */
void genome_map__generate_output(GenomeMapType* genome_map, FILE* output, FILE* seq_output) {
  int i, j, ri = 0;
#define SEGMENT_SIZE  5
#define SEGMENT_END   "__________ "
#define SEGMENT_IN    "           "
#define SEGMENT_START "%9d| "
#define INDENT 11
#define SEQUENCE_DISTANCE 1
#define SUMMARY_SEPARATOR "  |  "
#define IDENTICAL_MARKER  "  "
#define MISMATCH_MARKER   "<>"
  CODE_TYPE* read_sequence;
  CODE_TYPE* read_sequence_rev;

  int read_stack_size = genome_map->reads_db->read_len +  GENOME_MAP_INDELS + 1;

  /* Holds the indices of the currently displayed reads */
  int *read_indices;
  /* Holds the position in the reads */
  int *read_positions;
  int *read_indel_pos;

  ReferenceDBType* ref_db = genome_map->ref_db;
  ReadsDBType* reads_db = genome_map->reads_db;
  HitMapType* map       = genome_map->hitmap;

  int ref_color, reads_consensus_color, ref_position;;

  CODE_TYPE crt_result_base = ref_db->first_base;
  int result_len = 0;

  int displayed_read_width =
	  /* distance between sequences                        */ SEQUENCE_DISTANCE +
	  /* the color aligned at the crt reference position   */ 1 +
	  /* space for 'colors' inserted in the read, if any   */ map->indel_count;
  int displayed_line_len =
	  /* information about reference position, at the beginning    */ INDENT +
	  /* space for displaying the reference color at that position */ 1 +
	  /* distance between sequences                                */ 2* SEQUENCE_DISTANCE +
	  /* for each read, displayed_read_width (see above)           */ read_stack_size * displayed_read_width +
	  /* separator between the mapped reads and their "consensus"  */ 5 +
	  /* again, the crt position in the reference sequence         */ 1 +
	  /* mark if the consensus equals the reference or not         */ 2 +
	  /* display the mapped color                                  */ 1 +
	  /* and the inserted colors, if any                           */ map->indel_count +
	  /* a \n and a \0                                             */ 2;

  int pos = 0, k;
  char* displayed_line;

  /* Memory allocations */
  SAFE_FAILURE__ALLOC(read_sequence_rev, COMPRESSED_LEN(genome_map->reads_db->read_len), CODE_TYPE);
  /* --- */
  /* Holds the indices of the currently displayed reads */
  SAFE_FAILURE__ALLOC(read_indices, read_stack_size, int);
  /* Holds the position in the reads */
  SAFE_FAILURE__ALLOC(read_positions, read_stack_size, int);
  SAFE_FAILURE__ALLOC(read_indel_pos, read_stack_size, int);
  /* --- */
  SAFE_FAILURE__ALLOC(displayed_line, displayed_line_len, char);
  /* Done memory allocations */

  displayed_line[displayed_line_len - 2] = '\n';
  displayed_line[displayed_line_len - 1] = '\0';


  fprintf(seq_output, ">%s\n", ref_db->name);

  if (output) {
	  fprintf(output, "\n\n>%s\n", ref_db->name);
  }
  for (i = 0; i < read_stack_size; ++i) {
    read_indices[i] = i;
    read_positions[i] = -i;
  }
  for (i = 0; i < ref_db->size; ++i) {
    /* cleanup displayed line */
    memset(displayed_line, ' ', displayed_line_len - 2);
    sprintf(displayed_line, SEGMENT_START, i);
    pos = INDENT;
    ref_color = (int)NTH_CODE(ref_db->sequence, i);
    // print the color in the reference
    displayed_line[pos] = COLOR_CODE[(int)ref_color];
    pos += 3; /* 1 color + 2 spaces */
    // for each read, display the corresponding color(s), and advance in the read sequence accordingly
    for (j = 0; j < read_stack_size; ++j) {
      ref_position = read_indices[j];
      if (ref_position < ref_db->size) {
        ri = genome_map__get_read_idx_at(genome_map, ref_position);
        read_sequence = (((ri >= 0) && (ri < reads_db->size)) ? (reads_db->reads[ri].sequence) : NULL);
      } else {
        read_sequence = NULL;
      }
      if (read_sequence) {
        for (k = 0; k < MAP_DETAIL_SIZE; ++k) {
    	  if (map->map[ri][k].ref_position == ref_position &&  map->map[ri][k].sense & ALIGNMENT_SENSE_REVERSE) {
    		sequence__reverse_compressed((read_sequence), read_sequence_rev, reads_db->read_len);
    		read_sequence = read_sequence_rev;
    		break;
    	  }
        }
      }
      genome_map__display_read_pos__colors(read_sequence, &read_positions[j],
          read_sequence ? map->map[ri]->indel_positions: NULL, map->indel_count, &read_indel_pos[j],
          genome_map->g_map[ref_position],
          displayed_line + (pos + j * displayed_read_width));
      if (read_positions[j] >= reads_db->read_len) {
        read_positions[j] = i - ref_position - read_stack_size + 1;
        read_indices[j] += read_stack_size;
        read_indel_pos[j] = 0;
      }
    }
    // find and print the read consensus color
    pos += read_stack_size * displayed_read_width;
    reads_consensus_color =  genome_map__get_consensus_color(genome_map, i, -1, ref_color);
    sprintf(displayed_line + pos, "  |  %d%s%c", ref_color,
    	(ref_color == reads_consensus_color) ? "  " : MISMATCH_MARKER,
    	(reads_consensus_color == GENOME_MAP_NO_COLOR) ? '-' : COLOR_CODE[reads_consensus_color]);
    // further, compute and output the actual corresponding base in the resulted genome
    if (reads_consensus_color == GENOME_MAP_NO_COLOR) {
    	//deletion in the result
    	if (result_len == 0) {
    		//haven't started yet; see what base comes next
    		crt_result_base = TRANSFORM(crt_result_base, ref_color);
    	}
    }
//    if (reads_consensus_color < COLOR_CODE_COUNT) {
//    	if (result_len == 0) {
//    		// haven't started yet: print the first base and increment the counter
//    		fprintf(seq_output, "%c", BASE[(int)crt_result_base]);
//    		++result_len;
//    	}
//    	// compute and print the next base
//    	GET_AND_PRINT_BASE(seq_output, crt_result_base, reads_consensus_color);
//    }
    pos += 9;
    for (j = 0; j < map->indel_count; ++j) {
    	reads_consensus_color =  genome_map__get_consensus_color(genome_map, i, j, GENOME_MAP_NO_COLOR);
    	sprintf(displayed_line + pos, "%c", (reads_consensus_color == GENOME_MAP_NO_COLOR) ? ' ' : COLOR_CODE[reads_consensus_color]);
    	// compute and print the next base
//    	if (reads_consensus_color < COLOR_CODE_COUNT) {
//    		GET_AND_PRINT_BASE(seq_output, crt_result_base, reads_consensus_color);
//    	}
    }
    if (output) {
		fprintf(output, "%s", displayed_line);
		if (genome_map__get_read_idx_at(genome_map, i) >= 0) {
		  fprintf(output, "     %10d  ", genome_map->g_map[i]->read_idx);
		  display_compressed_code_sequence(genome_map->reads_db->reads[genome_map->g_map[i]->read_idx].sequence, genome_map->reads_db->read_len, output);
		} else {
		  fprintf(output, "\n");
		}
		if (output != stdout) {
			VERB_FILTER(VERBOSITY_MODERATE, display_progress(i, ref_db->size, 0))
		}
    }
  }
  if (output) {
	  fflush(output);
  }
  fprintf(seq_output, "\n");
  fflush(seq_output);
}

/**
 * Generates the CIGAR string and the right DNA sequence (no reading errors) from
 * the color sequence alignment.
 *
 * @param traceback a compressed representation of the alignment between the read and the reference
 * @param traceback_len the length of this alignment
 * @param ref_sequence pointer to the beginning of the reference fragment aligned to the read
 * @param offset the offset of the beginning
 * @param ref_start the first symbol of the reference sequence
 * @param seq_size
 * @param CIGAR_dest the CIGAR representation will be written here
 * @param SEQ_dest the DNA sequence will be written here
 */
inline void genome_map__traceback_to_CIGAR(const unsigned char* traceback, const int traceback_len,
		const CODE_TYPE* ref_sequence, const int offset, const char ref_start_symbol,
		char* CIGAR_dest, char* SEQ_dest) {
  int i, span = 1;
  char t_value;
#define STATE_MATCH  'M'
#define STATE_INSERT 'I'
#define STATE_DELETE 'D'
#define STATE_SNP    'S'
#define STATE_RE     'E'
#define STATE_NONE   0
  char crt_state = STATE_NONE;
  char prev_state = STATE_MATCH;
  char type;
  char* CIGAR_dest_pointer = CIGAR_dest;
  int SEQ_pointer = 0;
  int SEQ_len;
  CODE_TYPE code, ref_history = CODE_IDENTITY, read_history = CODE_IDENTITY;

  int ref_idx = offset;
  SEQ_dest[SEQ_pointer++] = CODE_IDENTITY;

  for (i = 0; i < traceback_len; ++i) {
    t_value = TRACEBACK_SEQ_GET(traceback, i);
    type = GET_PAIR_TYPE_SYMBOL(t_value);
    code = t_value & TRACEBACK_COLOR_MASK;
    /* SEQUENCE */
    switch (type) {
    case STATE_MATCH:
    case STATE_RE:
      //retrieve color from reference
      //SEQ_dest[SEQ_pointer++] = NTH_CODE(ref_sequence, ref_idx);
      // update the 'history' color of the reference
      ref_history = COMPOSE(ref_history,  NTH_CODE(ref_sequence, ref_idx));
      // compute the color to put in the read, and advance in the read:
      // it should be c such that:  read_history * c == ref_history
      // since each color is its own inverse, it follows that:  c = ref_history * read_history
      code = COMPOSE(ref_history, read_history);
      SEQ_dest[SEQ_pointer++] = code;
      // update the read history with the new color
      read_history = COMPOSE(read_history, code);
      // advance in the reference
      ref_idx++;
      break;
    case STATE_SNP:
      // copy color from read_color, and advance in both read and reference
      SEQ_dest[SEQ_pointer++] = code;
      read_history = COMPOSE(read_history,  code);
      ref_history = COMPOSE(ref_history,  NTH_CODE(ref_sequence, ref_idx));
      ref_idx++;
      break;
    case STATE_INSERT:
      // copy color from read and advance in the read
      SEQ_dest[SEQ_pointer++] = code;
      read_history = COMPOSE(read_history,  code);
      break;
    case STATE_DELETE:
      // skip color from reference
      ref_history = COMPOSE(ref_history,  NTH_CODE(ref_sequence, ref_idx));
      ref_idx++;
      break;
    default:
      break;
    }

    /* CIGAR */
    if ((type == STATE_DELETE) || (type == STATE_INSERT)) {
      crt_state = type;
    } else {
      crt_state = STATE_MATCH;
    }
    if (crt_state == prev_state) {
      ++span;
    } else {
      if (span) {
        sprintf(CIGAR_dest_pointer, "%d%c", span, prev_state);
        CIGAR_dest_pointer = CIGAR_dest + strlen(CIGAR_dest);
      }
      span = 1;
      prev_state = crt_state;
    }
  }

  if (span) {
    sprintf(CIGAR_dest_pointer, "%d%c", span, prev_state);
  }

  SEQ_len = SEQ_pointer;
  code = ref_start_symbol;
  // translate obtained color sequence using the reference letter
  for (SEQ_pointer = 0; SEQ_pointer < SEQ_len; ++SEQ_pointer) {
    code = TRANSFORM(code, SEQ_dest[SEQ_pointer]);
    SEQ_dest[SEQ_pointer] = BASE[code];
  }
}

/**
 * Outputs the mapped reads in the SAM fomat (http://samtools.sourceforge.net)
 * @param map
 * @param ref_db
 * @param reads_db
 * @param sam_output
 */
void genome_map__generate_SAM_output(GenomeMapType* genome_map, FILE* sam_output) {
  int i, k, ri = 0, ref_position;
  CODE_TYPE ref_symbol;
  ReadDataType* read_;

  ReferenceDBType* ref_db = genome_map->ref_db;
  ReadsDBType* reads_db   = genome_map->reads_db;
  HitMapType* map         = genome_map->hitmap;
  ref_symbol = ref_db->first_base;

  // Worst case: 1M1I1MII.... => 4*read length
  int CIGAR_max_size = reads_db->read_len << 2;
  char* CIGAR_STRING;
  int TRANSLATED_READ_max_size = reads_db->read_len + 2;
  char* TRANSLATED_READ_STRING;
  int alignment_quality;
  int flag = 0x0;

#define FLAG_EMPTY          0x0000
#define FLAG_PAIRED         0x0001 /* the read is paired in sequencing, no matter whether it is mapped in a pair */
#define FLAG_MAPPED_PAIRED  0x0002 /* the read is mapped in a proper pair (depends on the protocol, normally inferred during alignment) 1 */
#define FLAG_UNMAPPED       0x0004 /* the query sequence itself is unmapped */
#define FLAG_MATE_UNMAPPED  0x0008 /* the mate is unmapped */
#define FLAG_REVERSE        0x0010 /* strand of the query (0 for forward; 1 for reverse strand) */
#define FLAG_MATE_REVERSE   0x0020 /* strand of the mate */
#define FLAG_FIRST_MATE     0x0040 /* the read is the first read in a pair */
#define FLAG_SECOND_MATE    0x0080 /* the read is the second read in a pair */
#define FLAG_NOT_PRIMARY    0x0100 /* the alignment is not primary (a read having split hits may have multiple primary alignment records) */
#define FLAG_FAIL           0x0200 /* the read fails platform/vendor quality checks */
#define FLAG_DUPLICATE      0x0400 /* the read is either a PCR duplicate or an optical duplicate */

  if (!sam_output) {
    ERROR__(("No file name provided for SAM output."));
    return;
  }

  /* Memory allocations */
  SAFE_FAILURE__ALLOC(CIGAR_STRING, CIGAR_max_size, char);
  SAFE_FAILURE__ALLOC(TRANSLATED_READ_STRING, TRANSLATED_READ_max_size, char);

  /* --- */

  /* Display file header */
  /*
  Type           | Tag    | Description
  ---------------+--------+------------------------------------------------------------------------
  HD - header    | VN*    | File format version.
                 | SO     | Sort order. Valid values are: unsorted, queryname or coordinate.
                 | GO     | Group order (full sorting is not imposed in a group). Valid values are:
                 |        |   none, query or reference.
  ---------------+--------+------------------------------------------------------------------------
  SQ - Sequence  | SN*    | Sequence name. Unique among all sequence records in the ﬁle. The value
       dictionary|        | of this ﬁeld is used in alignment records.
                 | LN*    | Sequence length.
                 | AS     | Genome assembly identiﬁer. Refers to the reference genome assembly in
                 |        | an unambiguous form. Example: HG18.
                 | M5     | MD5 checksum of the sequence in the uppercase (gaps and space removed)
                 | UR     | URI of the sequence
                 | SP     | Species.
  ---------------+--------+------------------------------------------------------------------------
  RG - read group| ID*    | Unique read group identiﬁer. The value of the ID ﬁeld is used in the
                 |        | RG tags of alignment records.
                 | SM*    | Sample (use pool name where a pool is being sequenced)
                 | LB     | Library
                 | DS     | Description
                 | PU     | Platform unit (e.g. lane for Illumina or slide for SOLiD); should be a
                 |        | full, unambiguous identiﬁer
                 | PI     | Predicted median insert size (maybe different from the actual median
                 |        | insert size)
                 | CN     | Name of sequencing center producing the read.
                 | DT     | Date the run was produced (ISO 8601 date or date/time).
                 | PL     | Platform/technology used to produce the read.
  ---------------+--------+------------------------------------------------------------------------
  PG - Program   | ID*    | Program name
                 | VN     | Program version
                 | CL     | Command line
  ---------------+--------+------------------------------------------------------------------------
  CO - Comment   |        | One-line text comment
  */

#define DISPLAY_NAME(db) (db->name ? db->name : "UNKNOWN")
  fprintf(sam_output, "@HD\tVN:1.0\n");
  fprintf(sam_output, "@SQ\tSN:%s\tLN:%ld\n", DISPLAY_NAME(ref_db), (long)(ref_db->size + 1));
  fprintf(sam_output, "@RG\tID:%s\tSM:%s\n", DISPLAY_NAME(reads_db),  DISPLAY_NAME(reads_db));
  fprintf(sam_output, "@PG\tID:%s\tCL:", PROGRAM_NAME);
  for (i = 0; i < cmd_line_len; ++i) {
    fprintf(sam_output, "%s ", cmd_line[i]);
  }
  fprintf(sam_output, "\n");

  /* Display read */
  /*
  Field | Regular expression    | Range         | Description
  ------+-----------------------+---------------+--------------------------------------------------
  QNAME | ([^ \t\n\r])+         |               | Query pair NAME if paired; Query NAME if unpaired
  ------+-----------------------+---------------+--------------------------------------------------
  FLAG  | [0-9]+                | [0,2^16-1]    | bitwise FLAG
  ------+-----------------------+---------------+--------------------------------------------------
  RNAME | ([^ \t\n\r@-])+       |               | Reference sequence name
  ------+-----------------------+---------------+--------------------------------------------------
  POS   | [0-9]+                | [0,2^29-1]    | 1-based leftmost POSition/coordinate of the
        |                       |               | clipped sequence
  ------+-----------------------+---------------+--------------------------------------------------
  MAPQ  | [0-9]+                | [0,2^8-1]     | MAPping Quality (phred-scaled posterior probability
        |                       |               |  that the mapping position of this read is incorrect)
  ------+-----------------------+---------------+--------------------------------------------------
  CIGAR | ([0-9]+[MIDNSHP])+|\* |               | extended CIGAR string
  ------+-----------------------+---------------+--------------------------------------------------
  MRNM  | ([^ \t\n\r@])+        |               | Mate Reference sequence NaMe; “=” if the same as <RNAME>
  ------+-----------------------+---------------+--------------------------------------------------
  MPOS  | [0-9]+                | [0,2^29-1]    | 1-based leftmost Mate POSition of the clipped sequence
  ------+-----------------------+---------------+--------------------------------------------------
  ISIZE | -?[0-9]+              | [-2^29, 2^29] | inferred Insert SIZE
  ------+-----------------------+---------------+--------------------------------------------------
  SEQ   | [acgtnACGTN.=]+|\*    |               | query SEQuence; “=” for a match to the reference;
        |                       |               | n/N/. for ambiguity; cases are not maintained
  ------+-----------------------+---------------+--------------------------------------------------
  QUAL  | [!-~]+|\*             | [0,93]        | query QUALity; ASCII-33 gives the Phred base quality
  ------+-----------------------+---------------+--------------------------------------------------
  TAG   | [A-Z][A-Z0-9]         |               | TAG
  ------+-----------------------+---------------+--------------------------------------------------
  VTYPE | [AifzH]               |               | Value TYPE
  ------+-----------------------+---------------+--------------------------------------------------
  VALUE | [^\t\n\r]             |               | match <VTYPE> (space allowed)
  */

  // 2_66_251_F3     16      merged  2241562 255     35M     *       0       0
  // CTACATTGCAGTTGAGTACGGCAGTCATCTAATAA     #%!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%     NM:i:1  CS:Z:T0330322312121303131221012131031032        CQ:Z:###################################

  for (i = 0; i < GENOME_MAP_SIZE; ++i) {
    flag = FLAG_EMPTY;
    ri = genome_map__get_read_idx_at(genome_map, i);
    read_ = (((ri >= 0) && (ri < reads_db->size)) ? &(reads_db->reads[ri]) : NULL);
    if (read_) {
      ref_position = i + 1;

    // for (k = 0; k < MAP_DETAIL_SIZE; ++k) {
  	//   if (map->map[ri][k].ref_position == i) {
      if ((k = genome_map__get_read_score_rank_at(genome_map, i)) != UNMAPPED /*&& map->map[ri][k].ref_position == i*/) {
          if (map->map[ri][k].sense & ALIGNMENT_SENSE_REVERSE) {
            flag |= FLAG_REVERSE;
          }
          alignment_quality = map->map[ri][k].adjusted_score; // TODO: better computation
          memset(CIGAR_STRING, 0x0, CIGAR_max_size);
          memset(TRANSLATED_READ_STRING, 0x0, TRANSLATED_READ_max_size);

          genome_map__traceback_to_CIGAR(map->map[ri][k].traceback, map->map[ri][k].traceback_seq_len,
                                         ref_db->sequence + COMPRESSED_IDX(i), COMPRESSED_OFFSET(i), ref_symbol,
                                         CIGAR_STRING, TRANSLATED_READ_STRING);

    	  /* Read name, flag, reference name, mapping position, alignment quality, CIGAR string */
    	  fprintf(sam_output, "%s\t%d\t%s\t%d\t%d\t%s\t", read_->info, flag, ref_db->name, ref_position, alignment_quality, CIGAR_STRING);
    	  /* The mate information (mate reference name, mate position, inserted size) is not supported yet (TODO) */
    	  fprintf(sam_output, "=\t0\t0\t");
    	  /* Sequence (translated) and quality */

    	  /* Reversing is not needed for sam...
    	  if (flag & FLAG_REVERSE) {
    		  REVERSE_COMPLEMENT(TRANSLATED_READ_STRING);
    	  } */

          fprintf(sam_output, "%s\t", TRANSLATED_READ_STRING);
          int q;
          if (reads_db->reads[ri].quality) {
             if (map->map[ri][k].sense & ALIGNMENT_SENSE_REVERSE) {
               for (q = reads_db->read_len - 1; q >= 0; --q) {
                 fprintf(sam_output, "%c", 33 + READ_QUALITY_LEVEL_UPPER_BOUNDS[NTH_QUAL(reads_db->reads[ri].quality, q)]);
               }
               fprintf(sam_output, "%c", 33 + READ_QUALITY_LEVEL_UPPER_BOUNDS[reads_db->reads[ri].first_qual]);
             } else {
               fprintf(sam_output, "%c", 33 + READ_QUALITY_LEVEL_UPPER_BOUNDS[reads_db->reads[ri].first_qual]);
               for (q = 0; q < reads_db->read_len; ++q) {
                 fprintf(sam_output, "%c", 33 + READ_QUALITY_LEVEL_UPPER_BOUNDS[NTH_QUAL(reads_db->reads[ri].quality, q)]);
               }
             }
          } else {
             for (q = 0; q < reads_db->read_len + 1; ++q) {
               fprintf(sam_output, "%c", 33 + MAX_READ_QUALITY);
             }
          }
    	  fprintf(sam_output, "\n");
          //break;
    	//}
      }
    }
    ref_symbol = TRANSFORM(ref_symbol, NTH_CODE(ref_db->sequence, i));
  }
  fflush(sam_output);
}

/**
 * Quick display of pairs ref_position: read_index for each position with a corresponding mapped read
 */
void genome_map__display(const GenomeMapType* genome_map) {
  int i;
  for (i = 0; i < GENOME_MAP_SIZE; ++i) {
    if (genome_map__get_read_idx_at(genome_map, i) != UNMAPPED) {
    	int ri = genome_map->g_map[i]->read_idx;
    	int rank = genome_map->g_map[i]->score_rank;
    	int ref_pos = genome_map->hitmap->map[ri][rank].ref_position;
    	int tb_len = genome_map->hitmap->map[ri][rank].traceback_seq_len;
    	int score = genome_map->hitmap->map[ri][rank].score;
    	int a_score = genome_map->hitmap->map[ri][rank].adjusted_score;
      printf("%d: %d, %d (%d -- %d, %d, %d) \n", i, ri, rank, ref_pos, tb_len, score, a_score);
    }
  }
}

/**
 * Free the memory occupied by the genome map
 */
void genome_map__destroy(GenomeMapType* genome_map) {
  int i;
  for (i = 0; i < GENOME_MAP_SIZE; ++i) {
    if (genome_map->g_map[i]) {
#ifndef COMPRESSED_COLOR_COUNTERS
      if (genome_map->g_map[i]->aligned_positions_freqs) {
    	  free(genome_map->g_map[i]->aligned_positions_freqs);
      }
#endif
      if (genome_map->g_map[i]->inserted_positions_freqs) {
#ifndef COMPRESSED_COLOR_COUNTERS
    	int j;
	    for(j = 0; j < GENOME_MAP_INDELS; ++j) {
          free(genome_map->g_map[i]->inserted_positions_freqs[j]);
        }
#endif
        free(genome_map->g_map[i]->inserted_positions_freqs);
      }
      free(genome_map->g_map[i]);
    }
  }
  free(genome_map->g_map);
  free(genome_map->sorted_read_indices);
}
