#include "hit_map.h"
#include "util.h"

/**
 * Creates a map.
 * @param size
 * @return The address of the created map
 */
HitMapType* hit_map__create(int size, int indel_count) {
  // for each read, memorize MAP_DETAIL_SIZE best scores
  HitMapType* map;
  SAFE_FAILURE__ALLOC(map, 1, HitMapType);
  hit_map__init(map, size, indel_count);
  return map;
}

void hit_map__init(HitMapType* map, int size, int indel_count) {
  int i;
  map->size = size;
  map->indel_count = indel_count;
  map->mapped = 0;
  // for each read, memorize MAP_DETAIL_SIZE best scores
  SAFE_FAILURE__ALLOC(map->map, size, HitInfoType*);
  for (i = 0; i < size; ++i) {
    SAFE_FAILURE__ALLOC(map->map[i], MAP_DETAIL_SIZE, HitInfoType);
  }
  hit_map__clean(map);
}

inline void hit_map__clean(HitMapType* map) {
  int i, j;
  for (i = 0; i < map->size; ++i) {
    for(j = 0; j < MAP_DETAIL_SIZE; ++j) {
      map->map[i][j].score = MIN_SCORE;
      map->map[i][j].ref_position = UNMAPPED;
      map->map[i][j].indel_positions = NULL;
    }
  }
}
/**
 * Updates a map with a new alignment
 * @param map The map to update
 * @param read_id The aligned read's index in the reads database
 * @param ref_position The position where the aligned part of the reference starts
 * @param alignment The obtained alignment
 * @return The rank of the alignment if its score was good enough to include in the map
 * (between 0 and MAP_DETAIL_SIZE exclusive), or -1 otherwise.
 */
int hit_map__update(HitMapType* map, const int read_id, const int ref_position, const AlignmentType* alignment, int alignment_sense) {
  HitInfoType* info = map->map[read_id];
  int i, j, p;
  ScoreType score = alignment__get_score(alignment, alignment->best_i, alignment->best_j);
  /* If the update is successful, and the read wasn't mapped before, to the number of mapped reads */
  int mapped = (info[0].score == MIN_SCORE) ? 1 : 0;
  i = MAP_DETAIL_SIZE - 1;
  while (i >= 0 && score > info[i].score) {
    --i;
  }
  if (i ==  MAP_DETAIL_SIZE - 1) {
    // score is smaller, so no update
    return -1;
  }
 for (j = MAP_DETAIL_SIZE-1; j > i+1; --j) {
    info[j].score = info[j-1].score;
    info[j].ref_position = info[j-1].ref_position;
    info[j].sense = info[j-1].sense;
    if (info[j-1].indel_positions) {
      if (!info[j].indel_positions){
        SAFE_FAILURE__ALLOC(info[j].indel_positions, map->indel_count, short);
      }
      memcpy(info[j].indel_positions, info[j-1].indel_positions, map->indel_count*sizeof(short));
    }
    info[j].traceback_seq_len = info[j-1].traceback_seq_len;
    if (info[j-1].traceback) {
      if (!info[j].traceback){
        SAFE_FAILURE__ALLOC(info[j].traceback, TRACEBACK_SEQ_SIZE(info[j].traceback_seq_len), unsigned char);
      }
      memcpy(info[j].traceback, info[j-1].traceback,  TRACEBACK_SEQ_SIZE(info[j].traceback_seq_len)*sizeof(unsigned char));
    }
  }
  info[j].score = score;
  info[j].ref_position = (ref_position + alignment->best_j0);
  info[j].sense = (alignment_sense & ALIGNMENT_SENSE_REVERSE ? (-1) : 1);
  if (!info[j].indel_positions) {
    SAFE_FAILURE__ALLOC(info[j].indel_positions, map->indel_count, short);
  }
  memcpy(info[j].indel_positions, alignment->read_indel_positions, map->indel_count*sizeof(short));
  SAFE_FAILURE__ALLOC(info[j].traceback, TRACEBACK_SEQ_SIZE(alignment->traceback_seq_len), unsigned char);
  info[j].traceback_seq_len = alignment->traceback_seq_len;
  for (p = 0; p < info[j].traceback_seq_len; ++p) {
	  TRACEBACK_SEQ_SET(info[j].traceback, p, TRACEBACK_SEQ_GET(alignment->traceback, (alignment->traceback_seq_len - p - 1)));
  }
  //memcpy(info[j].traceback, alignment->traceback, TRACEBACK_SEQ_SIZE(alignment->traceback_seq_len));
  map->mapped += mapped;
  return j;
}
/**
 * Obtain the score for a read, when mapped on a given position
 * @param map
 * @param read_id The read index
 * @param ref_position The position of interest in the reference genome
 * @param rank The rank of that score will be placed at this address
 * @return The score
 */
inline int hit_map__score_for(HitMapType* map, int read_id, int ref_position, int *rank) {
  int i;
  for (i = 0; i < MAP_DETAIL_SIZE; ++i) {
    if ( map->map[read_id][i].ref_position == ref_position) {
      *rank = i;
      return map->map[read_id][i].score;
    }
  }
  *rank = MAP_DETAIL_SIZE;
  return MIN_SCORE;
}

/**
 * Quick display of the successfully matched reads: index, position on the reference, score, alignment length, color sequence, alignment summary
 */
void hit_map__display(const HitMapType* map, const ReferenceDBType* ref_db, const ReadsDBType* reads_db) {
  int i,j;
  for (i = 0; i < map->size; ++i) {
    for (j = 0; j < MAP_DETAIL_SIZE && map->map[i][j].ref_position != UNMAPPED; ++j) {
        printf("ri: %d  refpos: %d  score: %d  adj_score: %d  len: %d\n", i, map->map[i][j].ref_position, map->map[i][j].score, map->map[i][j].adjusted_score, map->map[i][j].traceback_seq_len);
        display_compressed_code_sequence(reads_db->reads[i].sequence, reads_db->read_len, stdout);
        int k;
        TRACEBACK_DISPLAY(map->map[i][j].traceback, map->map[i][j].traceback_seq_len, k, stdout);
        printf("\n");
    }
  }
}

/**
 * Cleanup
 */
void hit_map__destroy(HitMapType* map) {
  int i, j;
  for (i = 0; i < map->size; ++i) {
    for (j = 0; j < MAP_DETAIL_SIZE; ++j) {
      if (map->map[i][j].indel_positions) {
          free(map->map[i][j].indel_positions);
      } else {
    	  break;
      }
      if (map->map[i][j].traceback) {
          free(map->map[i][j].traceback);
      } else {
    	  break;
      }
    }
  }
}

