#include "reference_data.h"
#include <ctype.h>


int IGNORE_LOWERCASE = 0;

/**
 * Load reference sequence
 */
inline int load_reference_sequence(FILE* ref_in, ReferenceDBType* db) {
  char line[MAX_LINE_LEN];
  int line_len, i=0, masked_region_start = -1;
  char c;
  CODE_TYPE c1, c2;
  MaskedRegionType* crt_masked_region = db->masked_regions;

  db->size = 0;
  db->name = NULL;

  // browse the file once
  /* how big is the reference? */
  /* Expected format: fasta (one 'header' line, followed by the sequence) */
  // skip header...
  if (fgets(line, MAX_LINE_LEN, ref_in) && get_input_line_type(line) == LINE_TYPE_SEQ_HEADER) {
	  line_len = strlen(line) - 2; /* without the '>' and the '\n' */
	  if (line_len > 0) {
		  SAFE_FAILURE__ALLOC(db->name, line_len + 1, char); /* + 1 space for the '\0' */
		  strncpy(db->name, line+1, line_len);
	  }
	  else {
		  return RETURN_INPUT_ERR;
	  }
  } else {
	  return RETURN_INPUT_ERR;
  }

  i = ftell(ref_in);

  if (fgets(line, MAX_LINE_LEN, ref_in) && get_input_line_type(line) != LINE_TYPE_SEQ_HEADER) {
	  // this is the first line of the sequence
	  line_len = strlen(line);

	  /* ignore the newline */
	  while (line_len > 0 && isspace(line[line_len - 1])) {
		  --line_len;
	  }
	  if (line_len <= 0) {
		  return 0;
	  }
	  /* without the first base */
	  db->size = line_len - 1;

	  db->first_base = BASE_SYMBOL_TO_CODE(line[0]);

	  while (fgets(line, MAX_LINE_LEN, ref_in) && get_input_line_type(line) != LINE_TYPE_SEQ_HEADER) {
		  line_len = strlen(line);
		  /* ignore the newline */
		  while (line_len > 0 && isspace(line[line_len - 1])) {
		    --line_len;
		  }
		  db->size +=line_len;
	  }
	  /* Allocate */
	  SAFE_FAILURE__ALLOC(db->sequence, COMPRESSED_LEN(db->size), CODE_TYPE);

	  /* Load sequence */
	  fseek(ref_in, i, SEEK_SET);
	  c = (IGNORE_LOWERCASE ? fgetc(ref_in) : toupper(fgetc(ref_in)));
	  c1 = BASE_SYMBOL_TO_CODE(c);
	  i = 0;
	  while ((c = (IGNORE_LOWERCASE ? fgetc(ref_in) : toupper(fgetc(ref_in)))) != EOF) {
		  if (isspace(c)) {
			  continue;
		  }
		  if (c == SEQ_HEADER_MARKER) {
			  /* started to read the next sequence; getting out */
			  fseek(ref_in, -1, SEEK_CUR);
			  break;
		  }
		  if (!IS_VALID_BASE_SYMBOL(c)) {
			  c2 = c1; /* masked regions, eg 'NNNNNNNNN' cannot be coded; instead, GNNNNNNNN is transformed in GGGGGGGGG */
			  if (masked_region_start == -1) {
				  /* beginning of a new masked region */
				  masked_region_start = i;
			  }
		  } else {
			  c2 = BASE_SYMBOL_TO_CODE(c);
			  if (masked_region_start != -1) {
				  /* a masked region ends here */
				  if (crt_masked_region == NULL) {
					  /* this is the first one */
					  SAFE_FAILURE__ALLOC(db->masked_regions, 1, MaskedRegionType);
					  crt_masked_region = db->masked_regions;
				  } else {
					  SAFE_FAILURE__ALLOC(crt_masked_region->next, 1, MaskedRegionType);
					  crt_masked_region = crt_masked_region->next;
				  }
				  crt_masked_region->start = masked_region_start;
				  crt_masked_region->end = i;
				  masked_region_start = -1;
			  }
		  }
		  TO_NTH_CODE(db->sequence, i, COLOR(c1, c2));
		  c1 = c2;
		  ++i;
	  }
  }

  return db->size;
}

/**
 * Load reference sequence
 */
int load_reference_db(const char* reference_filename, ReferenceDBType** db) {
  char line[MAX_LINE_LEN];
  FILE* ref_in = fopen(reference_filename, "r");
  int i=0;
  int sequence_count = 0;

  if (!ref_in) {
    ERROR__(("Failed to read input file %s (reference sequence).\n", reference_filename));
    return RETURN_INPUT_ERR;
  }
  // browse the file once
  /* how many sequences are there?  */
  /* Expected format: multifasta (one 'header' line, followed by the sequence) */
  // So count headers...
  while (fgets(line, MAX_LINE_LEN, ref_in)) {
	  if (get_input_line_type(line) == LINE_TYPE_SEQ_HEADER) {
		  ++sequence_count;
	  }
  }

  /* Allocate */
  SAFE_FAILURE__ALLOC(*db, sequence_count, ReferenceDBType);

  /* Go back to the beginning and start loading */
  fseek (ref_in, 0, SEEK_SET);

  for (i = 0; i < sequence_count; ++i) {
	  if (load_reference_sequence(ref_in, &((*db)[i])) <= 0) {
		  /* something strange happened, stop loading and return what we have */
		  sequence_count = i;
		  break;
	  }
  }
  fclose(ref_in);

  return sequence_count;
}

/**
 * Destroys a reference database
 */
void clear_reference_db(ReferenceDBType* db) {
  MaskedRegionType* tmp;
  while (db->masked_regions != NULL) {
	  tmp = db->masked_regions;
	  db->masked_regions = db->masked_regions->next;
	  free(tmp);
  }
  if (db->name) {
    free(db->name);
  }
  free(db->sequence);
}

